# -*- coding: utf-8 -*-
"""
Created on Tue Sep 27 19:01:26 2022

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

from simpandas import SimSeries
from pandas import Series
import numpy as np
from unyts import convert, units

# assert default parameters
assert type(SimSeries()) is SimSeries
assert len(SimSeries()) == 0
assert SimSeries().units == {}
assert SimSeries().verbose is False
assert SimSeries().index_units is None
assert SimSeries().name_separator == ':'
assert SimSeries().intersection_character == '∩'
assert SimSeries()._auto_append_ is False
assert SimSeries()._operate_per_name_ is False
assert SimSeries()._transposed_ is False
assert str(SimSeries().dtype) == 'object'
assert SimSeries()._class is SimSeries


# prepare SimSeries
s0 = Series(
    data=[1/12, 2/12, 0.25, 0.5, 1.0, 1.5, 2.0],
    )
s1 = Series(
    data=[1/12, 2/12, 0.25, 0.5, 1.0, 1.5, 2.0],
    name='room'
    )
ss1 = SimSeries(
    data=[1/12, 2/12, 0.25, 0.5, 1.0, 1.5, 2.0],
    units='ft',
    )
ss2 = SimSeries(
    data=[1, 2, 3, 6, 12, 18, 24],
    units='yd',
    name='garden',
    )
ss3 = SimSeries(
    data=[1/12, 2/12, 0.25, 0.5, 1.0, 1.5, 2.0],
    units='ft',
    name='cookware:knifes'
    )
ss4 = SimSeries(
    data=[1, 2, 3, 6, 12, 18, 24],
    units='in',
    name='cookware:forks',
    )
ss5 = SimSeries(
    data=[1, 2, 3, 6, 12, 18, 24],
    units='in',
    name='table:forks',
    )
ss6 = SimSeries(
    data=[1, 2, 3, 6, 12, 18, 24],
    units='m',
    name='pool',
    )
s2 = ss4.as_pandas() + 2
m = units(1.0, 'm')
z = units(0.0, 'mm')
y = units(1.0, 'yd')
f = units(1.0, 'ft')
i = units(3.0, 'in')
t = units(3.5, 'h')


# call the series to return the values
assert ss1(5) == 1.5
assert ss2(2) == 3
assert (ss1() == ss1.values).all()


# get item
assert ss1[5] == units(1.5, 'ft')
assert ss2[0] == units(1, 'yd')
assert (ss2[1:3] == SimSeries(data=[2, 3, 6], units='in', index=[1, 2, 3])).all()
assert (ss2[1:3] == SimSeries(data=ss2.to_pandas().loc[1:3], units='in')).all()

# operations with SimSeries
## add
### add SimSeries
test = ss1+ss1
assert (test == ss1.as_pandas() + ss1.as_pandas()).all()
assert test.name is None
assert test.units == 'ft'

test = ss1+ss2
assert (test == ss1.to_numpy() + convert(ss2.to_numpy(), ss2.units, ss1.units)).all()
assert test.name == ss2.name
assert test.units == ss1.units

test = ss2+ss1
assert (test == ss2.to_numpy() + convert(ss1.to_numpy(), ss1.units, ss2.units)).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = ss2+ss6
assert (test == ss2.to_numpy() + convert(ss6.to_numpy(), ss6.units, ss2.units)).all()
assert test.name == ss2.name + '+' + ss6.name
assert test.units == ss2.units

test = ss6+ss2
assert (test == ss6.to_numpy() + convert(ss2.to_numpy(), ss2.units, ss6.units)).all()
assert test.name == ss6.name + '+' + ss2.name
assert test.units == ss6.units

test = ss3+ss4
assert (test == ss3.to_numpy() + convert(ss4.to_numpy(), ss4.units, ss3.units)).all()
assert test.name == 'cookware:knifes+forks'
assert test.units == ss3.units

test = ss4+ss3
assert (test == ss4.to_numpy() + convert(ss3.to_numpy(), ss3.units, ss4.units)).all()
assert test.name == 'cookware:forks+knifes'
assert test.units == ss4.units

test = ss4+ss5
assert (test == ss4.to_numpy() + convert(ss5.to_numpy(), ss5.units, ss4.units)).all()
assert test.name == 'cookware+table:forks'
assert test.units == ss4.units

test = ss5+ss4
assert (test == ss5.to_numpy() + convert(ss4.to_numpy(), ss4.units, ss5.units)).all()
assert test.name == 'table+cookware:forks'
assert test.units == ss5.units

### add Series
test = ss1+s0
assert (test == ss1.as_pandas() + s0).all()
assert test.name is None
assert test.units == ss1.units

test = ss2+s0
assert (test == ss2.as_pandas() + s0).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = ss2+s1
assert (test == ss2.as_pandas() + s1).all()
assert test.name == ss2.name + '+' + s1.name
assert test.units == ss2.units

test = s0+ss1
assert (test == ss1.as_pandas() + s0).all()
assert test.name is None
assert test.units == ss1.units

test = s0+ss2
assert (test == ss2.as_pandas() + s0).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = s1+ss2
assert (test == ss2.as_pandas() + s1).all()
assert test.name == s1.name + '+' + ss2.name
assert test.units == ss2.units

### add int
test = ss2 + 0
assert (test == ss2).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = 0 + ss2
assert (test == ss2).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = ss2 + 5
assert (test == ss2.to_numpy()+5).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = 5 + ss2
assert (test == 5+ss2.to_numpy()).all()
assert test.name == ss2.name
assert test.units == ss2.units

### add float
test = ss2 + 5.0
assert (test == ss2.to_numpy()+5.0).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = 5.0 + ss2
assert (test == 5.0+ss2.to_numpy()).all()
assert test.name == ss2.name
assert test.units == ss2.units

### add Unyt
test = ss2 + y
assert (test == ss2.to_numpy()+1).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = y + ss2
assert (test == 1+ss2.to_numpy()).all()
assert test.name == ss2.name
assert test.units == y.units

test = ss4 + y
assert (test == ss4.to_numpy()+1*12*3).all()
assert test.name == ss4.name
assert test.units == ss4.units

test = y + ss4
assert (test == 1+ss4.to_numpy()/12/3).all()
assert test.name == ss4.name
assert test.units == y.units

test = f + ss4
assert (test == f.values + ss4.to_numpy()/12).all()
assert test.name == ss4.name
assert test.units == f.units

test = ss4 + f
assert (test == ss4.to_numpy() + f.values*12).all()
assert test.name == ss4.name
assert test.units == ss4.units

test = m + ss3
assert (test == m.values + ss3.to_numpy()*0.3048).all()
assert test.name == ss3.name
assert test.units == m.units

test = ss3 + m
expected = (ss3.to_numpy() + m.values/0.3048).round(6)
assert (abs(test.to_numpy() - expected) < 1e-6).all()
assert test.name == ss3.name
assert test.units == ss3.units



## sub
### sub SimSeries
test = ss1-ss1
assert (test == ss1.as_pandas() - ss1.as_pandas()).all()
assert test.name is None
assert test.units == 'ft'

test = ss1-ss2
assert (test == ss1.to_numpy() - convert(ss2.to_numpy(), ss2.units, ss1.units)).all()
assert test.name == ss2.name
assert test.units == ss1.units

test = ss2-ss1
assert (test == ss2.to_numpy() - convert(ss1.to_numpy(), ss1.units, ss2.units)).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = ss2-ss6
assert (test == ss2.to_numpy() - convert(ss6.to_numpy(), ss6.units, ss2.units)).all()
assert test.name == ss2.name + '-' + ss6.name
assert test.units == ss2.units

test = ss6-ss2
assert (test == ss6.to_numpy() - convert(ss2.to_numpy(), ss2.units, ss6.units)).all()
assert test.name == ss6.name + '-' + ss2.name
assert test.units == ss6.units

test = ss3-ss4
assert (test == ss3.to_numpy() - convert(ss4.to_numpy(), ss4.units, ss3.units)).all()
assert test.name == 'cookware:knifes-forks'
assert test.units == ss3.units

test = ss4-ss3
assert (test == ss4.to_numpy() - convert(ss3.to_numpy(), ss3.units, ss4.units)).all()
assert test.name == 'cookware:forks-knifes'
assert test.units == ss4.units

test = ss4-ss5
assert (test == ss4.to_numpy() - convert(ss5.to_numpy(), ss5.units, ss4.units)).all()
assert test.name == 'cookware-table:forks'
assert test.units == ss4.units

test = ss5-ss4
assert (test == ss5.to_numpy() - convert(ss4.to_numpy(), ss4.units, ss5.units)).all()
assert test.name == 'table-cookware:forks'
assert test.units == ss5.units

### sub Series
test = ss1-s0
assert (test == ss1.as_pandas() - s0).all()
assert test.name is None
assert test.units == ss1.units

test = ss2-s0
assert (test == ss2.as_pandas() - s0).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = ss2-s1
assert (test == ss2.as_pandas() - s1).all()
assert test.name == ss2.name + '-' + s1.name
assert test.units == ss2.units

test = s0-ss1
assert (test == ss1.as_pandas() - s0).all()
assert test.name is None
assert test.units == ss1.units

test = s0-ss2
assert (test == s0 - ss2.as_pandas()).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = s1-ss2
assert (test == s1 - ss2.as_pandas()).all()
assert test.name == s1.name + '-' + ss2.name
assert test.units == ss2.units

### sub int
test = ss2 - 0
assert (test == ss2).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = 0 - ss2
assert (test == ss2.neg()).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = ss2 - 5
assert (test == ss2.to_numpy()-5).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = 5 - ss2
assert (test == 5-ss2.to_numpy()).all()
assert test.name == ss2.name
assert test.units == ss2.units

### sub float
test = ss2 - 5.0
assert (test == ss2.to_numpy()-5.0).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = 5.0 - ss2
assert (test == 5.0-ss2.to_numpy()).all()
assert test.name == ss2.name
assert test.units == ss2.units

### sub Unyt
test = ss2 - y
assert (test == ss2.to_numpy()-1).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = y - ss2
assert (test == 1-ss2.to_numpy()).all()
assert test.name == ss2.name
assert test.units == y.units

test = ss4 - y
assert (test == ss4.to_numpy()-1*12*3).all()
assert test.name == ss4.name
assert test.units == ss4.units

test = y - ss4
assert (test == 1-ss4.to_numpy()/12/3).all()
assert test.name == ss4.name
assert test.units == y.units

test = f - ss4
assert (test == f.values - ss4.to_numpy()/12).all()
assert test.name == ss4.name
assert test.units == f.units

test = ss4 - f
assert (test == ss4.to_numpy() - f.values*12).all()
assert test.name == ss4.name
assert test.units == ss4.units

test = m - ss3
assert (test == m.values - ss3.to_numpy()*0.3048).all()
assert test.name == ss3.name
assert test.units == m.units

test = ss3 - m
expected = (ss3.to_numpy() - m.values/0.3048).round(6)
assert (abs(test.to_numpy() - expected) < 1e-6).all()
assert test.name == ss3.name
assert test.units == ss3.units



## mul
### mul SimSeries
test = ss1*ss1
assert (test == ss1.as_pandas() * ss1.as_pandas()).all()
assert test.name is None
assert test.units == 'ft2'

test = ss1*ss2
assert (test == ss1.to_numpy() * convert(ss2.to_numpy(), ss2.units, ss1.units)).all()
assert test.name == ss2.name
assert test.units == ss1.units + '2'

test = ss2*ss1
assert (test == ss2.to_numpy() * convert(ss1.to_numpy(), ss1.units, ss2.units)).all()
assert test.name == ss2.name
assert test.units == ss2.units + '2'

test = ss2*ss6
assert (test == ss2.to_numpy() * convert(ss6.to_numpy(), ss6.units, ss2.units)).all()
assert test.name == ss2.name + '*' + ss6.name
assert test.units == ss2.units + '2'

test = ss6*ss2
assert (test == ss6.to_numpy() * convert(ss2.to_numpy(), ss2.units, ss6.units)).all()
assert test.name == ss6.name + '*' + ss2.name
assert test.units == ss6.units + '2'

test = ss3*ss4
assert (test == ss3.to_numpy() * convert(ss4.to_numpy(), ss4.units, ss3.units)).all()
assert test.name == 'cookware:knifes*forks'
assert test.units == ss3.units + '2'

test = ss4*ss3
assert (test == ss4.to_numpy() * convert(ss3.to_numpy(), ss3.units, ss4.units)).all()
assert test.name == 'cookware:forks*knifes'
assert test.units == ss4.units + '2'

test = ss4*ss5
assert (test == ss4.to_numpy() * convert(ss5.to_numpy(), ss5.units, ss4.units)).all()
assert test.name == 'cookware*table:forks'
assert test.units == ss4.units + '2'

test = ss5*ss4
assert (test == ss5.to_numpy() * convert(ss4.to_numpy(), ss4.units, ss5.units)).all()
assert test.name == 'table*cookware:forks'
assert test.units == ss5.units + '2'

### mul Series
test = ss1*s0
assert (test == ss1.as_pandas() * s0).all()
assert test.name is None
assert test.units == ss1.units

test = ss2*s0
assert (test == ss2.as_pandas() * s0).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = ss2*s1
assert (test == ss2.as_pandas() * s1).all()
assert test.name == ss2.name + '*' + s1.name
assert test.units == ss2.units

test = s0*ss1
assert (test == ss1.as_pandas() * s0).all()
assert test.name is None
assert test.units == ss1.units

test = s0*ss2
assert (test == ss2.as_pandas() * s0).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = s1*ss2
assert (test == ss2.as_pandas() * s1).all()
assert test.name == s1.name + '*' + ss2.name
assert test.units == ss2.units

### mul int
test = ss2 * 0
assert (test == 0).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = 0 * ss2
assert (test == 0).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = ss2 * 1
assert (test == ss2).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = 1 * ss2
assert (test == ss2).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = ss2 * 5
assert (test == ss2.to_numpy()*5).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = 5 * ss2
assert (test == 5*ss2.to_numpy()).all()
assert test.name == ss2.name
assert test.units == ss2.units

### mul float
test = ss2 * 5.0
assert (test == ss2.to_numpy()*5.0).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = 5.0 * ss2
assert (test == 5.0*ss2.to_numpy()).all()
assert test.name == ss2.name
assert test.units == ss2.units

### mul Unyt
test = ss2 * y
assert (test == ss2.to_numpy()*1).all()
assert test.name == ss2.name
assert test.units == ss2.units + '2'

test = y * ss2
assert (test == 1*ss2.to_numpy()).all()
assert test.name == ss2.name
assert test.units == y.units + '2'

test = ss4 * y
assert (test == ss4.to_numpy()*1*12*3).all()
assert test.name == ss4.name
assert test.units == ss4.units + '2'

test = y * ss4
assert (test == 1*ss4.to_numpy()/12/3).all()
assert test.name == ss4.name
assert test.units == y.units + '2'

test = f * ss4
assert (test == f.values * ss4.to_numpy()/12).all()
assert test.name == ss4.name
assert test.units == f.units + '2'

test = ss4 * f
assert (test == ss4.to_numpy() * f.values*12).all()
assert test.name == ss4.name
assert test.units == ss4.units + '2'

test = m * ss3
expected = (m.values * ss3.to_numpy()*0.3048).round(6)
assert (test == expected).all()
assert test.name == ss3.name
assert test.units == m.units + '2'

test = ss3 * m
expected = (ss3.to_numpy() * m.values/0.3048).round(6)
assert (test - expected < 1e-6).all()
assert test.name == ss3.name
assert test.units == ss3.units + '2'

test = ss3 * t
assert (test == ss3.to_numpy() * t.values).all()
assert test.name == ss3.name
assert test.units == ss3.units + '*' + t.units

test = t * ss3
assert (test == t.values * ss3.to_numpy()).all()
assert test.name == ss3.name
assert test.units == t.units + '*' + ss3.units



## div
### div SimSeries
test = ss1/ss1
assert (test == ss1.as_pandas() / ss1.as_pandas()).all()
assert test.name is None
assert test.units == 'ft/ft'

test = ss1/ss2
assert (test == ss1.to_numpy() / convert(ss2.to_numpy(), ss2.units, ss1.units)).all()
assert test.name == ss2.name
assert test.units == ss1.units + '/' + ss1.units

test = ss2/ss1
assert (test == ss2.to_numpy() / convert(ss1.to_numpy(), ss1.units, ss2.units)).all()
assert test.name == ss2.name
assert test.units == ss2.units + '/' + ss2.units

test = ss2/ss6
assert (test == ss2.to_numpy() / convert(ss6.to_numpy(), ss6.units, ss2.units)).all()
assert test.name == ss2.name + '/' + ss6.name
assert test.units == ss2.units + '/' + ss2.units

test = ss6/ss2
assert (test == ss6.to_numpy() / convert(ss2.to_numpy(), ss2.units, ss6.units)).all()
assert test.name == ss6.name + '/' + ss2.name
assert test.units == ss6.units + '/' + ss6.units

test = ss3/ss4
assert (test == ss3.to_numpy() / convert(ss4.to_numpy(), ss4.units, ss3.units)).all()
assert test.name == 'cookware:knifes/forks'
assert test.units == ss3.units + '/' + ss3.units

test = ss4/ss3
assert (test == ss4.to_numpy() / convert(ss3.to_numpy(), ss3.units, ss4.units)).all()
assert test.name == 'cookware:forks/knifes'
assert test.units == ss4.units + '/' + ss4.units

test = ss4/ss5
assert (test == ss4.to_numpy() / convert(ss5.to_numpy(), ss5.units, ss4.units)).all()
assert test.name == 'cookware/table:forks'
assert test.units == ss4.units + '/' + ss4.units

test = ss5/ss4
assert (test == ss5.to_numpy() / convert(ss4.to_numpy(), ss4.units, ss5.units)).all()
assert test.name == 'table/cookware:forks'
assert test.units == ss5.units + '/' + ss5.units

### div Series
test = ss1/s0
assert (test == ss1.as_pandas() / s0).all()
assert test.name is None
assert test.units == ss1.units

test = ss2/s0
assert (test == ss2.as_pandas() / s0).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = ss2/s1
assert (test == ss2.as_pandas() / s1).all()
assert test.name == ss2.name + '/' + s1.name
assert test.units == ss2.units

test = s0/ss1
assert (test == s0 / ss1.as_pandas()).all()
assert test.name is None
assert test.units == ss1.units + '-1'

test = s0/ss2
assert (test == s0 / ss2.as_pandas()).all()
assert test.name == ss2.name
assert test.units == ss2.units + '-1'

test = s1/ss2
assert (test == s1 / ss2.as_pandas()).all()
assert test.name == s1.name + '/' + ss2.name
assert test.units == ss2.units + '-1'

### div int
test = 0 / ss2
assert (test == 0).all()
assert test.name == ss2.name
assert test.units == ss2.units + '-1'

test = ss2 / 1
assert (test == ss2).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = 1 / ss2
assert (test == 1/ss2.to_numpy()).all()
assert test.name == ss2.name
assert test.units == ss2.units + '-1'

test = ss2 / 5
assert (test == ss2.to_numpy()/5).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = 5 / ss2
expected = 5 / ss2.to_numpy()
assert (test - expected < 1e-6).all()
assert test.name == ss2.name
assert test.units == ss2.units + '-1'

### div float
test = ss2 / 5.0
assert (test == ss2.to_numpy()/5.0).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = 5.0 / ss2
expected = 5.0 / ss2.to_numpy()
assert (test - expected < 1e-6).all()
assert test.name == ss2.name
assert test.units == ss2.units + '-1'

### div Unyt
test = ss2 / y
assert (test == ss2.to_numpy()/1).all()
assert test.name == ss2.name
assert test.units == ss2.units + '/' + ss2.units

test = y / ss2
assert (test == y.value/ss2.to_numpy()).all()
assert test.name == ss2.name
assert test.units == ss2.units + '/' + ss2.units

test = ss2 / i
expected = ss2.to_numpy() / (i.value/12/3)
assert (test - expected < 1e-6).all()
assert test.name == ss2.name
assert test.units == ss2.units + '/' + ss2.units

test = i / ss2
expected = i.value / (ss2.to_numpy()*3*12)
assert (test == expected).all()
assert test.name == ss2.name
assert test.units == i.units + '/' + i.units

test = ss4 / y
assert (test == ss4.to_numpy()/y.value/12/3).all()
assert test.name == ss4.name
assert test.units == ss4.units + '/' + ss4.units

test = y / ss4
expected = y.value / (ss4.to_numpy()/12/3)
assert (test == expected).all()
assert test.name == ss4.name
assert test.units == y.units + '/' + y.units

test = f / ss4
expected = f.values / (ss4.to_numpy()/12)
assert (test == expected).all()
assert test.name == ss4.name
assert test.units == f.units + '/' + f.units

test = ss4 / f
expected = ss4.to_numpy() / (f.values*12)
assert (test == expected).all()
assert test.name == ss4.name
assert test.units == ss4.units + '/' + ss4.units

test = m / ss3
expected = m.values / (ss3.to_numpy()*0.3048)
assert (test - expected < 1e-6).all()
assert test.name == ss3.name
assert test.units == m.units + '/' + m.units

test = ss3 / m
expected = ss3.to_numpy() / (m.values/0.3048)
assert (test - expected < 1e-6).all()
assert test.name == ss3.name
assert test.units == ss3.units + '/' + ss3.units

test = ss3 / t
assert (test == ss3.to_numpy() / t.values).all()
assert test.name == ss3.name
assert test.units == ss3.units + '/' + t.units

test = t / ss3
expected = t.values / ss3.to_numpy()
assert (test - expected < 1e-6).all()
assert test.name == ss3.name
assert test.units == t.units + '/' + ss3.units



## floordiv
### floordiv SimSeries
test = ss1//s0
assert (test == ss1.as_pandas() // s0).all()
assert test.name is None
assert test.units == ss1.units

test = ss2//s0
assert (test == ss2.as_pandas() // s0).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = ss2//s1
assert (test == ss2.as_pandas() // s1).all()
assert test.name == ss2.name + '/' + s1.name
assert test.units == ss2.units

test = s0//ss1
assert (test == s0 // ss1.as_pandas()).all()
assert test.name is None
assert test.units == ss1.units + '-1'

test = s0//ss2
assert (test == s0 // ss2.as_pandas()).all()
assert test.name == ss2.name
assert test.units == ss2.units + '-1'

test = s1//ss2
assert (test == s1 // ss2.as_pandas()).all()
assert test.name == s1.name + '/' + ss2.name
assert test.units == ss2.units + '-1'

### floordiv int
test = 0 // ss2
assert (test == 0).all()
assert test.name == ss2.name
assert test.units == ss2.units + '-1'

test = ss2 // 1
assert (test == ss2).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = 1 // ss2
assert (test == 1//ss2.to_numpy()).all()
assert test.name == ss2.name
assert test.units == ss2.units + '-1'

test = ss2 // 5
assert (test == ss2.to_numpy()//5).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = 5 // ss2
expected = 5 // ss2.to_numpy()
assert (test - expected < 1e-6).all()
assert test.name == ss2.name
assert test.units == ss2.units + '-1'

### floordiv float
test = ss2 // 5.0
assert (test == ss2.to_numpy()//5.0).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = 5.0 // ss2
expected = 5.0 // ss2.to_numpy()
assert (test - expected < 1e-6).all()
assert test.name == ss2.name
assert test.units == ss2.units + '-1'

### floordiv Unyt
test = ss2 // y
assert (test == ss2.to_numpy()//1).all()
assert test.name == ss2.name
assert test.units == ss2.units + '/' + ss2.units

test = y // ss2
expected = y.value // ss2.to_numpy()
assert (test == expected).all()
assert test.name == ss2.name
assert test.units == ss2.units + '/' + ss2.units

test = ss2 // i
expected = ss2.to_numpy() // (i.value/12/3)
assert (test - expected < 1e-6).all()
assert test.name == ss2.name
assert test.units == ss2.units + '/' + ss2.units

test = i // ss2
expected = i.value // (ss2.to_numpy()*3*12)
assert (test == expected).all()
assert test.name == ss2.name
assert test.units == i.units + '/' + i.units

test = ss4 // y
expected = ss4.to_numpy() // (y.value*12*3)
assert (test == expected).all()
assert test.name == ss4.name
assert test.units == ss4.units + '/' + ss4.units

test = y // ss4
expected = y.value // (ss4.to_numpy()/12/3)
assert (test == expected).all()
assert test.name == ss4.name
assert test.units == y.units + '/' + y.units

test = f // ss4
expected = f.values // (ss4.to_numpy()/12)
assert (test == expected).all()
assert test.name == ss4.name
assert test.units == f.units + '/' + f.units

test = ss4 // f
expected = ss4.to_numpy() // (f.values*12)
assert (test == expected).all()
assert test.name == ss4.name
assert test.units == ss4.units + '/' + ss4.units

test = m // ss3
expected = m.values // (ss3.to_numpy()*0.3048)
assert (test - expected < 1e-6).all()
assert test.name == ss3.name
assert test.units == m.units + '/' + m.units

test = ss3 // m
expected = ss3.to_numpy() // (m.values/0.3048)
assert (test - expected < 1e-6).all()
assert test.name == ss3.name
assert test.units == ss3.units + '/' + ss3.units

test = ss3 // t
assert (test == ss3.to_numpy() // t.values).all()
assert test.name == ss3.name
assert test.units == ss3.units + '/' + t.units

test = t // ss3
expected = t.values // ss3.to_numpy()
assert (test - expected < 1e-6).all()
assert test.name == ss3.name
assert test.units == t.units + '/' + ss3.units



## mod
### mod SimSeries
test = ss1%ss1
assert (test == ss1.as_pandas() % ss1.as_pandas()).all()
assert test.name is None
assert test.units == ss1.units

test = ss1%ss2
assert (test == ss1.to_numpy() % convert(ss2.to_numpy(), ss2.units, ss1.units)).all()
assert test.name == ss2.name
assert test.units == ss1.units

test = ss2%ss1
assert (test == ss2.to_numpy() % convert(ss1.to_numpy(), ss1.units, ss2.units)).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = ss2%ss6
assert (test == ss2.to_numpy() % convert(ss6.to_numpy(), ss6.units, ss2.units)).all()
assert test.name == ss2.name + '%' + ss6.name
assert test.units == ss2.units

test = ss6%ss2
assert (test == ss6.to_numpy() % convert(ss2.to_numpy(), ss2.units, ss6.units)).all()
assert test.name == ss6.name + '%' + ss2.name
assert test.units == ss6.units

test = ss3%ss4
assert (test == ss3.to_numpy() % convert(ss4.to_numpy(), ss4.units, ss3.units)).all()
assert test.name == 'cookware:knifes%forks'
assert test.units == ss3.units

test = ss4%ss3
assert (test == ss4.to_numpy() % convert(ss3.to_numpy(), ss3.units, ss4.units)).all()
assert test.name == 'cookware:forks%knifes'
assert test.units == ss4.units

test = ss4%ss5
assert (test == ss4.to_numpy() % convert(ss5.to_numpy(), ss5.units, ss4.units)).all()
assert test.name == 'cookware%table:forks'
assert test.units == ss4.units

test = ss5%ss4
assert (test == ss5.to_numpy() % convert(ss4.to_numpy(), ss4.units, ss5.units)).all()
assert test.name == 'table%cookware:forks'
assert test.units == ss5.units

### mod Series
test = ss1%s0
assert (test == ss1.as_pandas() % s0).all()
assert test.name is None
assert test.units == ss1.units

test = ss2%s0
assert (test == ss2.as_pandas() % s0).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = ss2%s1
assert (test == ss2.as_pandas() % s1).all()
assert test.name == ss2.name + '%' + s1.name
assert test.units == ss2.units

test = ss4%s2
assert (test == ss4.as_pandas() % s2).all()
assert test.name == 'cookware%cookware:forks'
assert test.units == ss4.units

test = s0%ss1
assert (test ==  s0 % ss1.as_pandas()).all()
assert test.name is None
assert test.units == {}

test = s0%ss2
assert (test == s0 % ss2.as_pandas()).all()
assert test.name == ss2.name
assert test.units == {}

test = s1%ss2
assert (test == s1 % ss2.as_pandas()).all()
assert test.name == s1.name
assert test.units == {}

test = s2%ss4
assert (test == s2 % ss4.as_pandas()).all()
assert test.name == s2.name
assert test.units == {}

### mod int
test = 0 % ss2
assert (test == 0).all()
assert test.name == ss2.name
assert test.units == {}

test = ss2 % 1
assert (test == ss2.to_numpy() % 1).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = 1 % ss2
assert (test == 1 % ss2.to_numpy()).all()
assert test.name == ss2.name
assert test.units == {}

test = ss2 % 5
assert (test == ss2.to_numpy()%5).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = 5 % ss2
assert (test == 5%ss2.to_numpy()).all()
assert test.name == ss2.name
assert test.units == {}

### mod float
test = ss2 % 5.0
assert (test == ss2.to_numpy()%5.0).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = 5.0 % ss2
assert (test == 5.0%ss2.to_numpy()).all()
assert test.name == ss2.name
assert test.units == {}

### mod Unyt
test = ss2 % y
assert (test == ss2.to_numpy()%1).all()
assert test.name == ss2.name
assert test.units == ss2.units

test = y % ss2
assert (test == y.value % ss2.to_numpy()).all()
assert test.name == y.name
assert test.units == y.units

test = ss4 % y
assert (test == ss4.to_numpy()%(1*12*3)).all()
assert test.name == ss4.name
assert test.units == ss4.units

test = y % ss4
assert (test == y.value % (ss4.to_numpy()/12/3)).all()
assert test.name == y.name
assert test.units == y.units

test = f % ss4
assert (test == f.values % (ss4.to_numpy()/12)).all()
assert test.name == f.name
assert test.units == f.units

test = ss4 % f
assert (test == ss4.to_numpy() % (f.values*12)).all()
assert test.name == ss4.name
assert test.units == ss4.units

test = m % ss3
expected = (m.values % (ss3.to_numpy()*0.3048))
assert (test.round(4).values == expected.round(4)).all()
assert test.name == m.name
assert test.units == m.units

test = ss3 % m
expected = (ss3.to_numpy() % (m.values/0.3048)).round(6)
assert (test - expected < 1e-6).all()
assert test.name == ss3.name
assert test.units == ss3.units

test = ss3 % t
assert (test == ss3.to_numpy() % t.values).all()
assert test.name == ss3.name
assert test.units == ss3.units

test = t % ss3
assert (test == t.values % ss3.to_numpy()).all()
assert test.name == t.name
assert test.units == t.units



# -*- coding: utf-8 -*-
"""
Created on Wed Sep 21 11:08:06 2022

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.1.2'
__release__ = 20230104
__all__ = ['write_schedule']

from simpandas.errors import OverwrittingError
import os
# from .._common.stringformat import date as strDate
from pandas import DataFrame

def write_schedule(self, path, units='FIELD', ControlMode=None, ShutStop=None):
    """
    export a eclipse style schedule file.

    Parameters
    ----------
    units : str or dict, optional
        a string 'FIELD', 'METRIC', LAB or PVT-M will convert the data to the corresponding eclipse simulator units system.
        a dictionary should contain desired units for all the columns to be converted. The default is None.
    ControlMode : str or dict, optional
        a string defining the control mode for the simulation:'ORAT','WRAT','GRAT'
        a dictionary with pairs of item:ControlModel for each item (well or group).
    ShutStop : str, optional
        a string 'OPEN, 'SHUT' or 'STOP' indicating what to do with the wells when their rate is zero.


    Returns
    -------
    None.
    """
    if os.path.isfile(path):
        raise OverwrittingError("The output file already exists:\n  '"+str(path)+"'")

    eclipseUnits0 = {'FIELD':{'OPR':'stb/day',  # Oil rate
                             'WPR':'stb/day',  # Water rate
                             'GPR':'Mscf/day',  # Gas rate
                             'LPR':'stb/day',  # Liquid rate
                             # 'RFR':'rb/day',  # Reservoir fluid volume rate
                             'BHP':'psia',  # BHP
                             'THP':'psia',  # THP
                             # 'WGPR': 'Mscf/day',  # Wet gas production rate
                             # 'TMR':'lb-M/day',  # Total molar rate
                             # 'SPR': 'stb/day',  # Steam production rate
                             },
                    'METRIC':{'OPR':'sm3/day',  # Oil rate
                             'WPR':'sm3/day',  # Water rate
                             'GPR':'sm3/day',  # Gas rate
                             'LPR':'sm3/day',  # Liquid rate
                             # 'RFR':'rm3/day',  # Reservoir fluid volume rate
                             'BHP':'barsa',  # BHP
                             'THP':'barsa',  # THP
                             # 'WGPR': 'sm3/day',  # Wet gas production rate
                             # 'TMR':'kg-M/day',  # Total molar rate
                             # 'SPR': 'sm3/day',  # Steam production rate
                             },
                    'LAB':{'OPR':'scc/hr',  # Oil rate
                             'WPR':'scc/hr',  # Water rate
                             'GPR':'scc/hr',  # Gas rate
                             'LPR':'scc/hr',  # Liquid rate
                             # 'RFR':'rcc/hr',  # Reservoir fluid volume rate
                             'BHP':'atma',  # BHP
                             'THP':'atma',  # THP
                             # 'WGPR': 'Mscf/day',  # Wet gas production rate
                             # 'TMR':'lb-M/day',  # Total molar rate
                             # 'SPR': 'stb/day',  # Steam production rate
                             },
                    'PVT-M':{'OPR':'sm3/day',  # Oil rate
                             'WPR':'sm3/day',  # Water rate
                             'GPR':'sm3/day',  # Gas rate
                             'LPR':'sm3/day',  # Liquid rate
                             # 'RFR':'rm3/day',  # Reservoir fluid volume rate
                             'BHP':'atma',  # BHP
                             'THP':'atma',  # THP
                             # 'WGPR': 'Mscf/day',  # Wet gas production rate
                             # 'TMR':'lb-M/day',  # Total molar rate
                             # 'SPR': 'stb/day',  # Steam production rate
                             }}

    # create dictionary for keyword parameters
    if units in eclipseUnits0:
        # eclipseUnits = {}
        # for each in eclipseUnits0[units]:
        #     for X in 'FGW':
        #         for H in ' H':
        #             eclipseUnits[X+each+H.strip()] = eclipseUnits0[units][each]
        # del eclipseUnits0
        # units = eclipseUnits
        units = { X+each+H.strip():eclipseUnits0[units][each] for each in eclipseUnits0[units] for X in 'FGW' for H in ' H' }
        del eclipseUnits0

    data = self.to(units).well_status().melt()

    indexName = data.index.name
    itemName = data.columns[2]

    if indexName is not None and itemName is not None and len(indexName)>0 and len(itemName)>0:
        data = data.reset_index().sort_values(by=[indexName,itemName,'attribute'],axis=0,ascending=True).set_index(indexName)
    else:
        data = data.melt().sort_index(axis=0,ascending=True)

    out = {'WCONHIST':DataFrame(index=data[itemName].unique(), columns=range(2,13)),
           'WCONINJH':DataFrame(index=data[itemName].unique(), columns=range(2,13)),
           'WCONPROD':DataFrame(index=data[itemName].unique(), columns=range(2,21)),
           'WCONINJE':DataFrame(index=data[itemName].unique(), columns=range(2,16))}

    if ShutStop is None:
        ShutStop = 'STOP'

    if type(ControlMode) is str:
        ControlMode = { item:ControlMode for item in data[itemName].unique() }

    itemCol = 2
    valueCol = 0
    attCol = 1

    curtime = ''
    lastime = None
    i = 0

    GORcriteria = 10

    outstr = []

    for i in range(len(data)):
        curtime = data.iloc[i].name
        if i == 0:
            lastime = curtime

        if lastime != curtime or i == len(data)-1:
            # prepare keywords to write
            out['WCONHIST']['keyword'] = 'WCONHIST'
            out['WCONINJH']['keyword'] = 'WCONINJH'
            out['WCONPROD']['keyword'] = 'WCONPROD'
            out['WCONINJE']['keyword'] = 'WCONINJE'
            keywords = out['WCONHIST'].append(out['WCONINJH']).append(out['WCONPROD']).append(out['WCONINJE'])
            keywords.dropna(axis=0, how='all',subset=range(2,13),inplace=True)

            prodh = keywords['keyword'] == 'WCONHIST'
            keywords.loc[prodh,2] = [ 'OPEN' if each else ShutStop for each in (keywords.loc[prodh,4:6].sum(axis=1) > 0) ]

            injeh = keywords['keyword'] == 'WCONINJH'
            keywords.loc[injeh,3] = [ 'OPEN' if each else ShutStop for each in (keywords.loc[injeh,4] > 0) ]

            prodf = keywords['keyword'] == 'WCONPROD'
            keywords.loc[prodf,2] = [ 'OPEN' if each else ShutStop for each in (keywords.loc[prodf,4:7].sum(axis=1) > 0) ]

            injef = keywords['keyword'] == 'WCONINJE'
            keywords.loc[injef,3] = [ 'OPEN' if each else ShutStop for each in (keywords.loc[injef,5:6].sum(axis=1) > 0) ]

            if ControlMode is None or item not in ControlMode:
                notnull = keywords.loc[prodh,[4,6,5]].notna()
                checkgor = DataFrame({4: keywords.loc[prodh,4].fillna(0) * GORcriteria > keywords.loc[prodh,6].fillna(0) , 5: (keywords.loc[prodh,5].notna()) & ((keywords.loc[prodh,4].isna()) & (keywords.loc[prodh,6].isna())) , 6: keywords.loc[prodh,4].fillna(0) * GORcriteria <= keywords.loc[prodh,6].fillna(0) })
                keywords.loc[prodh,3] = ((notnull*checkgor).astype(int) *   np.array(['ORAT','GRAT','WRAT']).reshape(1,-1)).sum(axis=1)
            else:
                keywords.loc[:,3] = [ ControlMode[item] for item in keywords.loc[:].index ]

# working on how to select from the several keyword lines for each well which is the correct keyword to write

            keywords['ranking'] = keywords.loc[:,4:20].count(axis=1)

            keywords.reset_index().sort_values(['index','keyword','ranking'],axis=0,ascending=[True,True,False]).groupby('index').first()[['keyword']]


            # write keywords
            for kw in keywords['keyword'].unique():
                outstr.append(kw)
                for i in range(len(keywords[keywords['keyword'] == kw])):
                    line = ' ' + ' '.join(map(str,keywords.reset_index().iloc[0].fillna('1*').drop('keyword').to_list())) + ' /'
                    outstr.append(line)
                outstr.append('/')
                outstr.append('\n')

            # write lastime
            if type(lastime) in (int,float):
                outstr.append('TSTEP')
                outstr.append(' ' + str(lastime) + ' /')
                outstr.append('/')
            else:
                outstr.append('DATES')
                outstr.append(' ' + dt.datetime(lastime).strftime("%d %d %Y %H:%M:%S") + ' /')
                outstr.append('/')
            outstr.append('\n')

            # reset out (keywords) dictionary
            out = {'WCONHIST':DataFrame(index=data[itemName].unique(), columns=range(2,13)),
                   'WCONINJH':DataFrame(index=data[itemName].unique(), columns=range(2,13)),
                   'WCONPROD':DataFrame(index=data[itemName].unique(), columns=range(2,21)),
                   'WCONINJE':DataFrame(index=data[itemName].unique(), columns=range(2,16))}

        # read the input table and put the values in the corresponding parameter of the keywords
        if data.iloc[i,attCol] == 'WSTATUS':  # oil production rate
            out['WCONHIST'].loc[data.iloc[i,itemCol],'Status'] = data.iloc[i,valueCol]
            out['WCONINJH'].loc[data.iloc[i,itemCol],'Status'] = data.iloc[i,valueCol]
            out['WCONPROD'].loc[data.iloc[i,itemCol],'Status'] = data.iloc[i,valueCol]
            out['WCONINJE'].loc[data.iloc[i,itemCol],'Status'] = data.iloc[i,valueCol]

        elif data.iloc[i,attCol] == 'WOPRH':  # oil production rate
            out['WCONHIST'].loc[data.iloc[i,itemCol],4] = data.iloc[i,valueCol]
        elif data.iloc[i,attCol] == 'WWPRH':  # water production rate
            out['WCONHIST'].loc[data.iloc[i,itemCol],5] = data.iloc[i,valueCol]
        elif data.iloc[i,attCol] == 'WGPRH':  # gas production rate
            out['WCONHIST'].loc[data.iloc[i,itemCol],6] = data.iloc[i,valueCol]
        elif data.iloc[i,attCol] == 'WVFPH':  # well VFP table number
            out['WCONHIST'].loc[data.iloc[i,itemCol],7] = data.iloc[i,valueCol]
            out['WCONINJH'].loc[data.iloc[i,itemCol],7] = data.iloc[i,valueCol]
        elif data.iloc[i,attCol] == 'WALQH':  # Artificial lift quantity
            out['WCONHIST'].loc[data.iloc[i,itemCol],8] = data.iloc[i,valueCol]
        elif data.iloc[i,attCol] == 'WTHPH':  # tubing head pressure (THP)
            out['WCONHIST'].loc[data.iloc[i,itemCol],9] = data.iloc[i,valueCol]
            out['WCONINJH'].loc[data.iloc[i,itemCol],6] = data.iloc[i,valueCol]
        elif data.iloc[i,attCol] == 'WBHPH':  # bottom hole pressure (BHP)
            out['WCONHIST'].loc[data.iloc[i,itemCol],10] = data.iloc[i,valueCol]
            out['WCONINJH'].loc[data.iloc[i,itemCol],5] = data.iloc[i,valueCol]
        elif data.iloc[i,attCol] == 'WWGPRH':  # wet gas production rate
            out['WCONHIST'].loc[data.iloc[i,itemCol],11] = data.iloc[i,valueCol]
        elif data.iloc[i,attCol] in ('WNPRH','WNLPRH'):  # NGL rate
            out['WCONHIST'].loc[data.iloc[i,itemCol],12] = data.iloc[i,valueCol]
        elif data.iloc[i,attCol] == 'WOIRH':
            out['WCONINJH'].loc[data.iloc[i,itemCol],4] = data.iloc[i,valueCol]
            out['WCONINJH'].loc[data.iloc[i,itemCol],2] = 'OIL'
        elif data.iloc[i,attCol] == 'WWIRH':
            out['WCONINJH'].loc[data.iloc[i,itemCol],4] = data.iloc[i,valueCol]
            out['WCONINJH'].loc[data.iloc[i,itemCol],2] = 'WATER'
        elif data.iloc[i,attCol] == 'WGIRH':
            out['WCONINJH'].loc[data.iloc[i,itemCol],4] = data.iloc[i,valueCol]
            out['WCONINJH'].loc[data.iloc[i,itemCol],2] = 'GAS'
        elif data.iloc[i,attCol] == 'WCTRL':
            out['WCONHIST'].loc[data.iloc[i,itemCol],3] = data.iloc[i,valueCol]
            out['WCONINJH'].loc[data.iloc[i,itemCol],12] = data.iloc[i,valueCol]
            out['WCONPROD'].loc[data.iloc[i,itemCol],3] = data.iloc[i,valueCol]
            out['WCONINJE'].loc[data.iloc[i,itemCol],4] = data.iloc[i,valueCol]

        elif data.iloc[i,attCol] == 'WOPR':  # oil production rate
            out['WCONPROD'].loc[data.iloc[i,itemCol],4] = data.iloc[i,valueCol]
        elif data.iloc[i,attCol] == 'WWPR':  # water production rate
            out['WCONPROD'].loc[data.iloc[i,itemCol],5] = data.iloc[i,valueCol]
        elif data.iloc[i,attCol] == 'WGPR':  # gas production rate
            out['WCONPROD'].loc[data.iloc[i,itemCol],6] = data.iloc[i,valueCol]
        elif data.iloc[i,attCol] == 'WLPR':  # liquid production rate
            out['WCONPROD'].loc[data.iloc[i,itemCol],7] = data.iloc[i,valueCol]
        elif data.iloc[i,attCol] == 'WVFP':  # Reservoir fluid volume rate
            out['WCONPROD'].loc[data.iloc[i,itemCol],8] = data.iloc[i,valueCol]
            # out['WCONINJH'].loc[data.iloc[i,itemCol],7] = data.iloc[i,valueCol]
        elif data.iloc[i,attCol] == 'WALQ':  # Artificial lift quantity
            out['WCONPROD'].loc[data.iloc[i,itemCol],12] = data.iloc[i,valueCol]
        elif data.iloc[i,attCol] == 'WVFP':  # well VFP table number
            out['WCONPROD'].loc[data.iloc[i,itemCol],11] = data.iloc[i,valueCol]
            out['WCONINJE'].loc[data.iloc[i,itemCol],9] = data.iloc[i,valueCol]
        elif data.iloc[i,attCol] == 'WTHP':  # tubing head pressure (THP)
            out['WCONPROD'].loc[data.iloc[i,itemCol],10] = data.iloc[i,valueCol]
            out['WCONINJE'].loc[data.iloc[i,itemCol],8] = data.iloc[i,valueCol]
        elif data.iloc[i,attCol] == 'WBHP':  # bottom hole pressure (BHP)
            out['WCONPROD'].loc[data.iloc[i,itemCol],9] = data.iloc[i,valueCol]
            out['WCONINJE'].loc[data.iloc[i,itemCol],7] = data.iloc[i,valueCol]
        elif data.iloc[i,attCol] == 'WWGPR':  # wet gas production rate
            out['WCONPROD'].loc[data.iloc[i,itemCol],13] = data.iloc[i,valueCol]
        elif data.iloc[i,attCol] in ('WNPR','WNLPR'):  # NGL rate
            out['WCONPROD'].loc[data.iloc[i,itemCol],20] = data.iloc[i,valueCol]
        # elif data.iloc[i,attCol] == 'WNPR':  # Total molar rate
        #     out['WCONPROD'].loc[data.iloc[i,itemCol],14] = data.iloc[i,valueCol]
        # elif data.iloc[i,attCol] == 'WNPR':  # Steam production rate
        #     out['WCONPROD'].loc[data.iloc[i,itemCol],15] = data.iloc[i,valueCol]
        # elif data.iloc[i,attCol] == 'WNPR':  # Calorific rate
        #     out['WCONPROD'].loc[data.iloc[i,itemCol],18] = data.iloc[i,valueCol]
        # elif data.iloc[i,attCol] == 'WNPR':  # Linearly combined rate
        #     out['WCONPROD'].loc[data.iloc[i,itemCol],19] = data.iloc[i,valueCol]
        elif data.iloc[i,attCol] == 'WOIR':  # Surface flow rate
            out['WCONINJE'].loc[data.iloc[i,itemCol],5] = data.iloc[i,valueCol]
            out['WCONINJE'].loc[data.iloc[i,itemCol],2] = 'OIL'
        elif data.iloc[i,attCol] == 'WWIR':  # Surface flow rate
            out['WCONINJE'].loc[data.iloc[i,itemCol],5] = data.iloc[i,valueCol]
            out['WCONINJE'].loc[data.iloc[i,itemCol],2] = 'WATER'
        elif data.iloc[i,attCol] == 'WGIR':  # Surface flow rate
            out['WCONINJE'].loc[data.iloc[i,itemCol],5] = data.iloc[i,valueCol]
            out['WCONINJE'].loc[data.iloc[i,itemCol],2] = 'GAS'
        # elif data.iloc[i,attCol] == 'WGIR':  # Vaporized oil concentration in the injected gas, or dissolved gas concentration in the injected oil
        #     out['WCONINJE'].loc[data.iloc[i,itemCol],10] = data.iloc[i,valueCol]
        # elif data.iloc[i,attCol] == 'WGIR':  # Thermal: ratio of gas volume to steam volume (C.W.E.) for a STEAM-GAS injector
        #     out['WCONINJE'].loc[data.iloc[i,itemCol],11] = data.iloc[i,valueCol]
        # elif data.iloc[i,attCol] == 'WGIR':  # Surface volume proportion of oil in a multiphase injecto
        #     out['WCONINJE'].loc[data.iloc[i,itemCol],12] = data.iloc[i,valueCol]
        # elif data.iloc[i,attCol] == 'WGIR':  # Surface volume proportion of water in a multiphase injecto
        #     out['WCONINJE'].loc[data.iloc[i,itemCol],13] = data.iloc[i,valueCol]
        # elif data.iloc[i,attCol] == 'WGIR':  # Surface volume proportion of gas in a multiphase injecto
        #     out['WCONINJE'].loc[data.iloc[i,itemCol],14] = data.iloc[i,valueCol]
        # elif data.iloc[i,attCol] == 'WGIR':  # Ratio of oil volume to steam volume (C.W.E.) for a STEAM-OIL injector (for use with steam-solvent injection).
        #     out['WCONINJE'].loc[data.iloc[i,itemCol],15] = data.iloc[i,valueCol]

        lastime = curtime


# -*- coding: utf-8 -*-
"""
Created on Sun Oct 11 11:14:32 2020

@author: Martin Carlos Araya
"""

__version__ = '0.83.0'
__release__ = 20230215
__all__ = ['SimSeries']

from pandas import Series, DataFrame, Index
from io import StringIO
from shutil import get_terminal_size
from pandas._config import get_option
import fnmatch
import numpy as np
import pandas as pd
import warnings

from unyts.converter import convertible as _convertible, convert_for_SimPandas as _converter
from unyts.operations import unit_product as _unit_product, unit_division as _unit_division, unit_base as _unit_base, \
    unit_power as _unit_power, unit_addition as _unit_addition
from unyts.dictionaries import unitless_names as _unitless_names
from unyts.helpers.common_classes import number
from unyts import units, is_Unit, Unit

from .basics import SimBasics
from .common.helpers import clean_axis as _clean_axis, string_new_name as _string_new_name
from .common.math import znorm as _znorm, minmaxnorm as _minmaxnorm, jitter as _jitter
from .common.slope import slope as _slope
from .indexer import _SimLocIndexer, _iSimLocIndexer
from .index import SimIndex

_SERIES_WARNING_MSG = """\
    You are passing unitless data to the SimSeries constructor. Currently,
    it falls back to returning a pandas Series. But in the future, we will start
    to raise a TypeError instead."""


def _simseries_constructor_with_fallback(data=None, index=None, units=None, **kwargs):
    """
    A flexible constructor for SimSeries._constructor, which needs to be able
    to fall back to a Series(if a certain operation does not produce
    units)
    """
    try:
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=_SERIES_WARNING_MSG,
                category=FutureWarning,
                module="SimPandas[.*]",
            )
            return SimSeries(data=data,
                             index=index,
                             units=units,
                             **kwargs)
    except TypeError:
        return Series(data=data,
                      index=index,
                      **kwargs)


class SimSeries(SimBasics, Series):
    """
    A Series object designed to store data with units.

    Parameters
    ----------
    data : array-like, dict, scalar value
        The data to store in the SimSeries.
    index : array-like or Index
        The index for the SimSeries.
    units : string or dictionary of units(optional)
        Can be any string, but only units acepted by the UnitConverter will
        be considered when doing arithmetic calculations with other SimSeries
        or SimDataFrames.

    kwargs
        Additional arguments passed to the Series constructor,
         e.g. ``name``.

    See Also
    --------
    SimDataFrame
    pandas.Series

    """
    _metadata = ['units',
                 'verbose',
                 'index_units',
                 'name_separator',
                 'intersection_character',
                 'spdLocator',
                 'spdiLocator',
                 'columns',
                 'meta',
                 'source_path',
                 '_auto_append_',
                 '_operate_per_name_',
                 '_transposed_',
                 '_reverse_',]

    def __init__(self,
                 data=None,
                 index=None,
                 columns=None,
                 units=None,
                 dtype=None,
                 name=None,
                 copy=False,
                 fastpath=False,
                 verbose=False,
                 index_name=None,
                 index_units=None,
                 name_separator=None,
                 intersection_character='∩',
                 auto_append=False,
                 operate_per_name=False,
                 transposed=False,
                 meta=None,
                 source_path=None,
                 *args, **kwargs):

        self.units = {}
        self.verbose = bool(verbose)
        self.index_units = None
        self.name_separator = None
        self.intersection_character = intersection_character if type(intersection_character) is str else '∩'
        self.spdLocator = _SimLocIndexer("loc", self)
        self.spdiLocator = _iSimLocIndexer("iloc", self)
        self.meta = meta
        self.source_path = source_path
        self._auto_append_ = bool(auto_append)
        self._operate_per_name_ = bool(operate_per_name)
        self._transposed_ = bool(transposed)
        self._reverse_ = kwargs['reverse'] if 'reverse' in kwargs else False

        # data validaton
        if isinstance(data, DataFrame) and len(data.columns) > 1:
            raise ValueError("'data' parameter can be an instance of DataFrame but must have only one column.")

        # get units from data if it is SimDataFrame or SimSeries
        if units is None or (type(units) in [list, dict] and len(units) == 0):
            if hasattr(data, 'get_units'):
                units = data.get_units()
        elif type(units) is dict and len(units) == 1:
            units, name = list(units.values())[0], list(units.keys())[0] if name is None else name
        elif type(units) is str:
            units = units.strip()

        # get name_separator
        if name_separator is None and hasattr(data, 'name_separator'):
            name_separator = data.name_separator
        elif name_separator is not None and type(name_separator) is str and len(name_separator.strip()) > 0:
            pass
        elif name_separator is False:
            name_separator = ''
        else:
            name_separator = ':'
        self.name_separator = name_separator

        # define default dtype
        if data is None and dtype is None:
            dtype = object

        # catch index units if index is instance of SimIndex
        if index_units is None and hasattr(index, 'units'):
            index_units = index.units

        # initialize pd.Series
        super().__init__(data=data, index=index, dtype=dtype, name=name, copy=copy, fastpath=fastpath)

        # get name
        if self.name is None or (type(self.name) is str and self.name.strip() == ''):
            if type(units) is dict and len(units) == 1:
                self.name = list(units.keys())[0]

        self.columns = Index([self.name]) if columns is None else columns

        # set units
        if units is not None:
            self.set_units(units)

        # get index_units
        if index_units is None:
            if self.index.name is not None and self.index.name in self.units:
                self.index_units = self.units[self.index.name]
            elif hasattr(data, 'index_units'):
                self.index_units = data.index_units.copy() if type(data.index_units) is dict else data.index_units
        elif type(index_units) is str:
            self.index_units = index_units
        else:
            raise TypeError("`index_units` must be a string.")

        # override index.name with index_name
        if index_name is not None:
            if type(self.units) is dict and self.index.name in self.units:
                self.units[index_name] = self.units[self.index.name]
            self.index.name = index_name

        # change pd.Index to SimIndex
        self.index = SimIndex(self.index, units=self.index_units)

    @property
    def _class(self):
        return SimSeries

    @property
    def _constructor(self):
        return _simseries_constructor_with_fallback

    @property
    def _constructor_expanddim(self):
        from simpandas.frame import SimDataFrame
        return SimDataFrame

    def to_pandas(self):
        return self.to_series()

    def as_pandas(self):
        return self.as_series()

    def to_series(self):
        return Series(self.copy())

    def as_series(self):
        return Series(self)

    def to_simseries(self):
        return self

    def as_simseries(self):
        return self

    def to_dataframe(self):
        return self.to_simdataframe().to_dataframe()

    def as_dataframe(self):
        return self.as_simdataframe().as_dataframe()

    def to_simdataframe(self):
        from .frame import SimDataFrame
        if type(self.units) is str:
            return SimDataFrame(data=self)
        elif type(self.units) is dict:
            return SimDataFrame(
                data=self.values.reshape(1, self.values.size),
                index=[self.name],
                columns=self.index,
                **self.params_)

    def as_simdataframe(self):
        return self.to_simdataframe()

    def __call__(self, key=None):
        """
        Returns the series values, a NumPy array or number without units.
        """
        if key is None:
            return self.values
        else:
            return self[key].values

    def __getitem__(self, key=None):
        from .frame import SimDataFrame
        if not hasattr(self.index, 'units'):
            self.index = SimIndex(self.index, units=self.index_units)
        def index_params_():
            params_ = self.params_.copy()
            params_['name'] = self.index.name
            params_['units'] = self.index.units
            params_['index_name'] = None
            params_['index_units'] = None
            return params_

        if key is None:
            return self
        elif type(key) is str and key not in self.index and key == self.name:
            return self
        elif type(key) is str and key == self.index.name:
            params_ = index_params_()
            return SimSeries(data=self.index.to_numpy(),
                             index=range(len(self.index)),
                             **params_)
        elif type(key) is list and key in [[], [self.name]]:
            return self.as_simdataframe()
        elif type(key) is list and key == [self.index.name]:
            params_ = index_params_()
            return SimDataFrame(data={self.index.name: self.index.to_numpy()},
                                index=range(len(self.index)),
                                **params_)
        else:
            try:
                result = self.loc[key]
            except (KeyError, pd.IndexingError):
                try:
                    result = self.iloc[key]
                except IndexError:
                    if type(key) is tuple:
                        try:
                            return self[list(key)]
                        except (IndexError, InvalidIndexError):
                            pass
                    try:
                        result = self.as_simdataframe()[key]
                    except:
                        raise KeyError("the requested Key is not a valid index or name: " + str(key))
        if isinstance(result, pd.Series) and len(result) == 1:
            if type(result.iloc[0]) in number:
                result = units(result.iloc[0], result.get_units_string())
            else:
                result = result.iloc[0]
        return result

    def __repr__(self) -> str:
        """
        Return a string representation for a particular Series, with Units.
        """

        # taken from Pandas Series
        buf = StringIO("")
        width, height = get_terminal_size()
        max_rows = (
            height
            if get_option("display.max_rows") == 0
            else get_option("display.max_rows")
        )
        min_rows = (
            height
            if get_option("display.max_rows") == 0
            else get_option("display.min_rows")
        )
        show_dimensions = get_option("display.show_dimensions")

        self.as_series().to_string(
            buf=buf,
            name=self.name,
            dtype=self.dtype,
            min_rows=min_rows,
            max_rows=max_rows,
            length=show_dimensions,
        )
        result = buf.getvalue()

        if type(self.units) is str:
            return result + ', units: ' + self.units
        elif type(self.units) is dict:
            result = result.split('\n')
            for n in range(len(result) - 1):
                keys = result[n] + ' '
                i, f = 0, 0
                while i < len(keys):
                    f = keys.index(' ', i)
                    key = keys[i:f]
                    if key == '...':
                        i = len(keys)
                        continue
                    while key not in self.index and f <= len(keys):
                        f = keys.index(' ', f + 1) if ' ' in keys[f + 1:] else len(keys) + 1
                        key = keys[i:f]
                    if key not in self.index:
                        i = len(keys)
                        continue
                    if key in self.units and self.units[key] is not None:
                        result[n] += '    ' + self.units[key].strip()
                    i = len(keys)
            result = '\n'.join(result)
            return '\n' + result
        else:
            return result

    def _arithmethic_operation(self, other, operation: str = None, level=None, fill_value=None, axis=0,
                               intersection_character=None):
        def _units_operation(a, b, operation):
            if operation in ['+', '-']:
                return _unit_addition(a, b)
            elif operation in ['*']:
                return _unit_product(a, b)
            elif operation in ['/', '//']:
                return _unit_division(a, b)
            elif operation in ['**']:
                return _unit_power(a, b)
            elif operation in ['%']:
                return a
            else:
                raise ValueError("Unknown operation")

        params_ = self.params_.copy()
        _products = ['*', '/', '//', '%']
        valid_operations = {# operator, pd.Series.method, proposed fill_value
            '+': [pd.Series.add, 'Addition', 0],
            '-': [pd.Series.sub, 'Subtraction', 0],
            '*': [pd.Series.mul, 'Product', 1],
            '/': [pd.Series.truediv, 'Division', None],
            '//': [pd.Series.floordiv, 'Floor Division', None],
            '%': [pd.Series.mod, 'Module', None],
            '**': [pd.Series.pow, 'Power', None],
            '^': [pd.Series.pow, 'Power', None]}
        assert operation in valid_operations
        intersection_character = operation if intersection_character is None else intersection_character
        op_method = valid_operations[operation][0]
        op_label = valid_operations[operation][1]
        fill_value = valid_operations[operation][1] if fill_value is True else fill_value

        # ensure self.index is SimIndex
        if not hasattr(self.index, 'units'):
            self.index = SimIndex(self.index, units=self.index_units)

        # both SimSeries
        if isinstance(other, SimSeries):
            if self.index.name is not None and other.index.name is not None and self.index.name != other.index.name:
                warnings.warn("indexes of both SimSeries are not of the same kind:\n   '" +
                              self.index.name + "' != '" + other.index.name + "'")

            # ensure other.index is SimIndex
            if not hasattr(other.index, 'units'):
                other.index = SimIndex(other.index, units=other.index_units)

            # convert other.index.units if required and possible
            if self.index.units == other.index.units:
                pass
            elif self.index.units not in _unitless_names and other.index.units not in _unitless_names and \
                    _convertible(other.index.units, self.index.units):
                other = other.index_to(self.index.units)

            if type(self.units) is str and type(other.units) is str:
                new_name = _string_new_name(
                    self._common_rename(other, intersection_character=intersection_character, return_names_dict_only=True),
                    intersection_character=intersection_character)
                params_['units'] = _units_operation(self.units, other.units, operation)
                if self.units == other.units:
                    result = op_method(self.as_pandas(), other.as_pandas(), level=level, fill_value=fill_value, axis=axis)
                elif _convertible(other.units, self.units):
                    other_c = _converter(other.as_pandas(), other.units, self.units, self.verbose)
                    result = op_method(self.as_pandas(), other_c, level=level, fill_value=fill_value, axis=axis)
                elif _convertible(self.units, other.units):
                    self_c = _converter(self.as_pandas(), self.units, other.units, self.verbose)
                    result = op_method(other.as_pandas(), self_c, level=level, fill_value=fill_value, axis=axis)
                    params_['units'] = _units_operation(other.units, self.units, operation)
                elif operation in _products and _convertible(_unit_base(other.units), _unit_base(self.units)):
                    other_c = _converter(other.as_pandas(), _unit_base(other.units), _unit_base(self.units),
                                         self.verbose)
                    result = op_method(self.as_pandas(), other_c, level=level, fill_value=fill_value, axis=axis)
                elif operation in _products and _convertible(_unit_base(self.units), _unit_base(other.units)):
                    self_c = _converter(self.as_pandas(), _unit_base(self.units), _unit_base(other.units), self.verbose)
                    result = op_method(other.as_pandas(), self_c, level=level, fill_value=fill_value, axis=axis)
                    params_['units'] = _units_operation(other.units, self.units, operation)
                else:
                    result = op_method(self.as_pandas(), other.as_pandas(), level=level, fill_value=fill_value, axis=axis)
                    if type(self.units) is str and type(other.units) is str:
                        params_['units'] = self.units + operation + other.units
                    elif type(self.units) is dict and len(self.units) == 1 and type(other.units) is str:
                        params_['units'] = self.get_units_string() + operation + other.units
                    elif type(other.units) is dict and len(other.units) == 1 and type(self.units) is str:
                        params_['units'] = self.units + operation + other.get_units_string()
                    elif type(self.units) is dict and len(self.units) == 1 and type(other.units) is dict and len(
                            other.units) == 1:
                        params_['units'] = self.get_units_string() + operation + other.get_units_string()
                    elif type(self.units) is dict and type(other.units) is dict:
                        params_['units'] = self.units.copy()
                        for k, u in other.units.items():
                            if k in params_['units']:
                                params_['units'][k] = params_['units'][k] + operation + u
                            else:
                                params_['units'][k] = u
                    else:
                        raise NotImplementedError(op_label + ' of SimSeries with different units is not implemented.')
                params_['name'] = new_name
                result = self._class(data=result, **params_)
            else:
                raise NotImplementedError

        # other is Pandas Series
        elif isinstance(other, Series):
            result = op_method(self.as_pandas(), other, level=level, fill_value=fill_value, axis=axis)
            new_name = _string_new_name(
                self._common_rename(self._class(other), intersection_character=intersection_character,
                                    return_names_dict_only=True),
                intersection_character=intersection_character)
            params_['name'] = new_name

        # other is int or float
        elif type(other) in (int, float, complex):
            result = op_method(self.as_pandas(), other, level=level, fill_value=fill_value, axis=axis)

        # other is instance of unyts
        elif is_Unit(other):
            if type(self.units) is str:
                if self._reverse_:
                    params_['units'] = _units_operation(other.units, self.units, operation)
                else:
                    params_['units'] = _units_operation(self.units, other.units, operation)
                if _convertible(other.unit, self.units):
                    result = op_method(self.as_pandas(), other.to(self.units).value, level=level, fill_value=fill_value, axis=axis)
                elif operation in _products:
                    result = op_method(self.as_pandas(), other.value, level=level, fill_value=fill_value, axis=axis)
                else:
                    raise NotImplementedError(op_label + " of SimSeries with not convertible Unyts is not implemented.")
            else:
                result = op_method(self.as_simdataframe().as_pandas(), other, level=level, fill_value=fill_value, axis=axis).as_simseries()

        # lets Pandas deal with other types, maintain units, dtype and name
        else:
            result = op_method(self.as_pandas(), other, level=level, fill_value=fill_value, axis=axis)

        if operation in ['//']:
            params_['dtype'] = result.dtype
        else:
            params_['dtype'] = self.dtype if result.astype(self.dtype).equals(result) else result.dtype
        self._reverse_ = False
        return self._class(data=result, **params_)

    def __add__(self, other):
        return self._arithmethic_operation(other, operation='+', fill_value=0)

    def __sub__(self, other):
        return self._arithmethic_operation(other, operation='-', fill_value=0)
    def __mul__(self, other):
        return self._arithmethic_operation(other, operation='*', fill_value=1)

    def __truediv__(self, other):
        return self._arithmethic_operation(other, operation='/', fill_value=None)

    def __floordiv__(self, other):
        return self._arithmethic_operation(other, operation='//', fill_value=None, intersection_character='/')

    def __mod__(self, other):
        return self._arithmethic_operation(other, operation='%', fill_value=None)
    def __pow__(self, other):
        return self._arithmethic_operation(other, operation='**', fill_value=None)

    def astype(self, dtype, copy=True, errors='raise'):
        params_ = self.params_.copy()
        params_['dtype'] = dtype
        return self._class(data=self.as_pandas().astype(dtype), **params_)

    def copy(self):
        if type(self.units) is dict:
            params_ = self.params_.copy()
            params_['units'] = self.units.copy()
            return SimSeries(data=self.as_pandas().copy(True), **params_)
        return SimSeries(data=self.as_pandas().copy(True), **self.params_)

    def convert(self, units):
        """
        returns the SimSeries converted to the requested units if possible, if not, returns the original values.
        """
        if isinstance(units, (Unit, SimSeries)):
            units = units.units
        if type(units) is str and type(self.units) is str:
            if _convertible(self.units, units):
                params_ = self.params_.copy()
                params_['units'] = units
                return self._class(data=_converter(self.as_pandas(), self.units, units, self.verbose), **params_)
            else:
                return self
        elif type(units) is str and type(self.units) is dict and len(set(self.units.values())) == 1:
            params_ = self.params_.copy()
            params_['units'] = units
            return self._class(data=_converter(self.as_pandas(), list(set(self.units.values()))[0], units, self.verbose), **params_)
        elif type(units) is dict:
            return self.to_simdataframe().convert(units).to_simseries()
        else:
            return self

    def drop(self, labels=None, axis=0, index=None, columns=None, level=None, inplace=False, errors='raise'):
        axis = _clean_axis(axis)
        if inplace:
            super().drop(labels=labels, axis=axis, index=index, columns=columns,
                         level=level, inplace=inplace, errors='errors')
        else:
            return SimSeries(data=self.as_pandas().drop(labels=labels, axis=axis, index=index, columns=columns,
                                                        level=level, inplace=inplace, errors='errors'),
                             **self.params_)

    def dropna(self, axis=0, inplace=False, how=None):
        axis = _clean_axis(axis)
        if inplace:
            super().dropna(axis=axis, inplace=inplace, how=how)
        else:
            return SimSeries(
                data=self.as_pandas().dropna(axis=axis, inplace=inplace, how=how),
                **self.params_)

    def drop_zeros(self, axis=None, inplace=False):
        """
        drop the rows where the values are zeross.
        """
        filt = self.zeros(axis=0)
        if inplace:
            self.drop(columns=filt[filt == True].index, inplace=True)
        else:
            return self.drop(columns=filt[filt == True].index, inplace=False)

    def filter(self, conditions=None, **kwargs):
        """
        Returns a filtered SimSeries based on conditions argument.

        To filter over the series simply define the
        condition:
            '>0'

        To set several conditions together the operatos 'and' and 'or'
        are accepted:
            '>0 and <1000'

        To filter only over the index set the condition directly:
            '>0'
        or use the key '.index' or '.i' to refer to the index of the SimSeries.

        To remove null values append '.notnull' to the column name:
            'NAME.notnull'
        To keep only null values append '.null' to the column name:
            'NAME'.null
        """
        from simpandas.common.filters import key_to_string

        return_string = False
        if 'return_string' in kwargs:
            return_string = bool(kwargs['return_string'])
        return_filter = False
        if 'return_filter' in kwargs:
            return_filter = bool(kwargs['return_filter'])
        return_frame = False
        if 'return_frame' in kwargs:
            return_frame = bool(kwargs['return_frame'])
        if 'return_series' in kwargs:
            return_frame = bool(kwargs['return_series'])
        if not return_filter and not return_string and ('return_series' not in kwargs or 'return_frame' not in kwargs):
            return_frame = True

        special_operation = ['.notnull', '.null', '.isnull', '.abs']
        numpy_operation = ['.sqrt', '.log10', '.log2', '.log', '.ln']
        pandas_aggregation = ['.any', '.all']
        pandas_agg = ''

        if type(conditions) is not str:
            if type(conditions) is not list:
                try:
                    conditions = list(conditions)
                except:
                    raise TypeError('conditions argument must be a string.')
            conditions = ' and '.join(conditions)

        conditions = conditions.strip() + ' '

        # find logical operators and translate to correct key
        and_or_not = False
        if ' and ' in conditions:
            conditions = conditions.replace(' and ', ' & ')
        if ' or ' in conditions:
            conditions = conditions.replace(' or ', ' | ')
        if ' not ' in conditions:
            conditions = conditions.replace(' not ', ' ~ ')
        if '&' in conditions:
            and_or_not = True
        elif '|' in conditions:
            and_or_not = True
        elif '~' in conditions:
            and_or_not = True

        # create Pandas compatible condition string
        filter_str = ' ' + '(' * and_or_not
        key = ''
        cond, oper = '', ''
        i = 0
        while i < len(conditions):

            # catch logital operators
            if conditions[i] in ['&', "|", '~']:
                filter_str, key, pandas_agg = key_to_string(filter_str, key, pandas_agg)
                filter_str = filter_str.rstrip()
                filter_str += ' )' + pandas_agg + ' ' + conditions[i] + '('
                pandas_agg = ''
                i += 1
                continue

            # catch enclosed strings
            if conditions[i] in ['"', "'", '[']:
                if conditions[i] in ['"', "'"]:
                    try:
                        f = conditions.index(conditions[i], i + 1)
                    except:
                        raise ValueError('wring syntax, closing ' + conditions[i] + ' not found in:\n   ' + conditions)
                else:
                    try:
                        f = conditions.index(']', i + 1)
                    except:
                        raise ValueError("wring syntax, closing ']' not found in:\n   " + conditions)
                if f > i + 1:
                    key = conditions[i + 1:f]
                    filter_str, key, pandas_agg = key_to_string(filter_str, key, pandas_agg)
                    i = f + 1
                    continue

            # pass blank spaces
            if conditions[i] == ' ':
                filter_str, key, pandas_agg = key_to_string(filter_str, key, pandas_agg)
                if len(filter_str) > 0 and filter_str[-1] != ' ':
                    filter_str += ' '
                i += 1
                continue

            # pass parenthesis
            if conditions[i] in ['(', ')']:
                filter_str, key, pandas_agg = key_to_string(filter_str, key, pandas_agg)
                filter_str += conditions[i]
                i += 1
                continue

            # catch conditions
            if conditions[i] in ['=', '>', '<', '!']:
                cond = ''
                f = i + 1
                while conditions[f] in ['=', '>', '<', '!']:
                    f += 1
                cond = conditions[i:f]
                if cond == '=':
                    cond = '=='
                elif cond in ['=>', '=<', '=!']:
                    cond = cond[::-1]
                elif cond in ['><', '<>']:
                    cond = '!='
                filter_str, key, pandas_agg = key_to_string(filter_str, key, pandas_agg)
                filter_str = filter_str.rstrip()
                filter_str += ' ' + cond
                i += len(cond)
                continue

            # catch operations
            if conditions[i] in ['+', '-', '*', '/', '%', '^']:
                oper = ''
                f = i + 1
                while conditions[f] in ['+', '-', '*', '/', '%', '^']:
                    f += 1
                oper = conditions[i:f]
                oper = oper.replace('^', '**')
                filter_str, key, pandas_agg = key_to_string(filter_str, key, pandas_agg)
                filter_str = filter_str.rstrip()
                filter_str += ' ' + oper
                i += len(oper)
                continue

            # catch other characters
            else:
                key += conditions[i]
                i += 1
                continue

        # clean up
        filter_str = filter_str.strip()
        # check missing key, means .index by default
        if filter_str[0] in ['=', '>', '<', '!']:
            filter_str = 'self.as_Series().index ' + filter_str
        elif filter_str[-1] in ['=', '>', '<', '!']:
            filter_str = filter_str + ' self.as_Series().index'
        # close last parethesis and aggregation
        filter_str += ' )' * bool(and_or_not + bool(pandas_agg)) + pandas_agg
        # open parenthesis for aggregation, if needed
        if not and_or_not and bool(pandas_agg):
            filter_str = '(' + filter_str

        ret_tuple = []

        if return_string:
            ret_tuple += [filter_str]
        filter_array = eval(filter_str)
        if return_filter:
            ret_tuple += [filter_array]
        if return_frame:
            ret_tuple += [self.as_pandas()[filter_array]]

        if len(ret_tuple) == 1:
            return ret_tuple[0]
        else:
            return tuple(ret_tuple)

    def rms(self, axis=0, **kwargs):
        return units((((self.as_pandas()) ** 2).mean(axis=axis, **kwargs)) ** 0.5,
                     self.get_UnitsString())

    def min(self, axis=0, **kwargs):
        return units(self.as_pandas().min(axis=axis, **kwargs), self.get_UnitsString())

    def max(self, axis=0, **kwargs):
        return units(self.as_pandas().max(axis=axis, **kwargs), self.get_UnitsString())

    def mean(self, axis=0, **kwargs):
        return units(self.as_pandas().mean(axis=axis, **kwargs), self.get_UnitsString())

    def mode(self, axis=0, **kwargs):
        return units(self.as_pandas().mode(axis=axis, **kwargs), self.get_UnitsString())

    def prod(self, axis=0, **kwargs):
        from unyts.operations import unit_base_power
        from unyts.units.unitless import unitless_names
        unit_base, unit_power = unit_base_power(self.get_UnitsString())
        prod_units = unit_base if unit_base in unitless_names else (unit_base + str(unit_power * len(self)))
        return units(self.as_pandas().prod(axis=axis, **kwargs), prod_units)

    def quantile(self, q=0.5, axis=0, **kwargs):
        return units(self.as_pandas().quantile(q, axis=axis, **kwargs), self.get_UnitsString())

    def sum(self, axis=0, **kwargs):
        return units(self.as_pandas().sum(axis=axis, **kwargs), self.get_UnitsString())

    def std(self, axis=0, **kwargs):
        return units(self.as_pandas().std(axis=axis, **kwargs), self.get_UnitsString())

    def var(self, axis=0, **kwargs):
        return units(self.as_pandas().var(axis=axis, **kwargs), self.get_UnitsString())

    def round(self, decimals=0, **kwargs):
        return units(self.as_pandas().round(decimals=decimals, **kwargs), self.get_UnitsString())

    def get_keys(self, pattern=None):
        """
        Will return a tuple of all the key names in case.

        If the pattern variable is different from None only keys
        matching the pattern will be returned; the matching is based
        on fnmatch():
            Pattern     Meaning
            *           matches everything
            ?           matches any single character
            [seq]       matches any character in seq
            [!seq]      matches any character not in seq

        """
        if pattern is not None and type(pattern) is not str:
            raise TypeError(
                'pattern argument must be a string.\nreceived ' + str(type(pattern)) + ' with value ' + str(pattern))
        if type(self.name) is str:
            keys = [self.name]
        else:
            keys = list(self.index)
        if pattern is None:
            return keys
        else:
            return list(fnmatch.filter(keys, pattern))

    def get_Units(self, items=None):
        return self.get_units()

    def get_units(self, items=None):
        """
        returns the units for the selected 'items' or for all the columns in the SimDataFrame.

        Parameters
        ----------
        items : str or list of str, optional
            Ignored, this parameter is kept for compatibility with SimDataFrame. The default is None.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        if self.units is None:
            units_dict = {self.name: 'unitless'}
        elif type(self.units) is str:
            units_dict = {self.name: self.units}
        elif type(self.units) is dict:
            units_dict = {each: (self.units[each] if each in self.units else 'unitless') for each in self.index }
        else:
            units_dict = self.units.copy() if type(self.units) is dict else {self.name: self.units}
        return units_dict

    def set_units(self, units, item=None):
        """
        This method can be used to define the units related to the values of a column (item).

        Parameters
        ----------
        units : str or list of str
            the units to be assigned
        item : str, optional
            The name of the column to apply the units.
            The default is None. In this case the unit

        Raises
        ------
        ValueError
            when units can't be applied.
        TypeError
            when units or item has the wrong format.

        Returns
        -------
        None.

        """
        if item is not None and type(item) in (str, int, float) and item not in self.columns and item not in self.index:
            raise ValueError("the required item '" + str(item) + "' is not in this SimSeries")

        if self.units is None or type(self.units) is str:
            if units is None:
                self.units = None
            elif type(units) is str:
                self.units = units.strip()
            elif type(units) is dict:
                old_units = self.units
                try:
                    self.units = {}
                    return self.set_units(units)
                except:
                    self.units = old_units
                    raise ValueError("not able to process dictionary of units.")
            else:
                raise TypeError("units must be a string.")

        elif type(self.units) is dict:
            if type(units) not in (str, dict) and hasattr(units, '__iter__'):
                if item is not None and type(item) not in (str, dict) and hasattr(item, '__iter__'):
                    if len(item) == len(units):
                        return self.set_units(dict(zip(item, units)))
                    else:
                        raise ValueError("both units and item must have the same length.")
                elif item is None:
                    if len(units) == len(self.columns):
                        return self.set_units(dict(zip(list(self.columns), units)))
                    else:
                        raise ValueError(
                            "units list must be the same length of columns in the SimSeries or must be followed by a list of items.")
                else:
                    raise TypeError("if units is a list, items must be a list of the same length.")
            elif type(units) is dict:
                for k, u in units.items():
                    try:
                        self.set_units(u, k)
                    except:
                        pass
            elif type(units) is str:
                if item is None:
                    self.units = units.strip()
                else:
                    if type(item) not in (str, dict) and hasattr(item, '__iter__'):
                        units = units.strip()
                        for i in item:
                            if i in self.units:
                                self.units[i] = units
                    elif type(item) is str:
                        if item in self.units:
                            self.units[item] = units
                        elif item in self.columns or item in self.index:
                            self.units[item] = units

            if item is None and len(self.columns) > 1:
                raise ValueError("More than one column in this SimSeries, item must not be None")
            elif item is None and type(units) is str and len(self.columns) == 1:
                return self.set_units(units, list(self.columns)[0])
            elif item is not None:
                if item in self.columns:
                    if units is None:
                        self.units[item] = None
                    elif type(units) is str:
                        self.units[item] = units.strip()
                    else:
                        raise TypeError("units must be a string.")
                if item == self.index.name:
                    self.index_units = units.strip()
                    self.units[item] = units.strip()
                elif item in self.index.names:
                    self.units[item] = units.strip()
                elif item in self.index:
                    self.units[item] = units.strip()

    def daily(self, agg='mean', datetime_index=False):
        """
        return a Series with a single row per day.
        index must be a date type.

        available grouping calculations are:
            first : keeps the first row per day
            last : keeps the last row per day
            max : returns the maximum value per year
            min : returns the minimum value per year
            mean or avg : returns the average value per year
            median : returns the median value per month
            std : returns the standard deviation per year
            sum : returns the summation of all the values per year
            count : returns the number of rows per year
        """
        return self.to_SimDataFrame().daily(agg=agg, datetime_index=datetime_index).to_simseries()

    def monthly(self, agg='mean', datetime_index=False):
        """
        return a dataframe with a single row per month.
        index must be a date type.

        available gropuing calculations are:
            first : keeps the fisrt row per month
            last : keeps the last row per month
            max : returns the maximum value per month
            min : returns the minimum value per month
            mean or avg : returns the average value per month
            median : returns the median value per month
            std : returns the standard deviation per month
            sum : returns the summation of all the values per month
            count : returns the number of rows per month

        datetimeIndex : bool
            if True the index will be converted to DateTimeIndex with Day=1 for each month
            if False the index will be a MultiIndex (Year,Month)
        """
        return self.to_SimDataFrame().monthly(agg=agg, datetime_index=datetime_index).to_simseries()

    def yearly(self, agg='mean', datetime_index=False):
        """
        return a dataframe with a single row per year.
        index must be a date type.

        available grouping calculations are:
            first : keeps the first row
            last : keeps the last row
            max : returns the maximum value per year
            min : returns the minimum value per year
            mean or avg : returns the average value per year
            median : returns the median value per month
            std : returns the standard deviation per year
            sum : returns the summation of all the values per year
            count : returns the number of rows per year

        datetimeIndex : bool
            if True the index will be converted to DateTimeIndex with Day=1 and Month=1 for each year
            if False the index will be a MultiIndex (Year,Month)
        """
        return self.to_SimDataFrame().yearly(agg=agg, datetime_index=datetime_index).to_simseries()

    def reindex(self, index=None, **kwargs):
        """
        wrapper for pandas.Series.reindex

        index : array-like, optional
            New labels / index to conform to, should be specified using keywords.
            Preferably an Index object to avoid duplicating data.
        """
        return SimSeries(data=self.to_pandas().reindex(index=index, **kwargs), **self.params_)

    def rename(self, index=None, *, axis=None, copy=True, inplace=False, level=None, errors='ignore', **kwargs):
        """
        wrapper of rename function from Pandas.

        Alter Series index labels or name.

        Function / dict values must be unique (1-to-1).
        Labels not contained in a dict / Series will be left as-is.
        Extra labels listed don’t throw an error.

        Alternatively, change Series.name with a scalar value.

        See the user guide for more.

        Parameters
        axis{0 or “index”}
        Unused. Accepted for compatibility with DataFrame method only.

        indexscalar, hashable sequence, dict-like or function, optional
        Functions or dict-like are transformations to apply to the index.
        Scalar or hashable sequence-like will alter the Series.name attribute.

        **kwargs
        Additional keyword arguments passed to the function. Only the “inplace” keyword is used.

        Returns
        Series or None
        Series with index labels or name altered or None if inplace=True.
        """

        # for compatibility with SimDataFrame
        if 'columns' in kwargs and index is None:
            index = kwargs['columns']
            del kwargs['columns']

        if type(index) is dict:
            if len(index) == 1 and list(index.keys())[0] not in self.index:
                return self.rename(list(index.values())[0], axis=axis, copy=copy, inplace=inplace, level=level,
                                   errors=errors)
            col_before = list(self.index)
            if inplace:
                super().rename(index=index, axis=axis, copy=copy, inplace=inplace, level=level, errors=errors)
                col_after = list(self.index)
            else:
                catch = super().rename(index=index, axis=axis, copy=copy, inplace=inplace, level=level, errors=errors)
                col_after = list(catch.index)

            new_units = {}
            for i in range(len(col_before)):
                new_units[col_after[i]] = self.units[col_before[i]]
            if inplace:
                self.units = new_units
                self.spdLocator = _SimLocIndexer("loc", self)
                return None
            else:
                catch.units = new_units
                catch.spdLocator = _SimLocIndexer("loc", catch)
                return catch
        elif type(index) is str:
            if inplace:
                self.name = index.strip()
                self.spdLocator = _SimLocIndexer("loc", self)
                return None
            else:
                catch = self.copy()
                catch.name = index
                catch.spdLocator = _SimLocIndexer("loc", catch)
                return catch

    def set_index(self, name):
        self.set_index_name(name)

    def set_index_units(self, units):
        if hasattr(units, 'units') and type(units.units) is str:
            units = units.units
        elif hasattr(units, 'unit') and type(units.unit) is str:
            units = units.unit
        if type(units) is str and len(units.strip()) > 0:
            self.index_units = units.strip()
        elif type(units) is dict and len(units) == len(self.index):
            self.index_units = units
        else:
            raise TypeError("`units` must be a string or a dictionary with pair key: units for each item in the index.")
        if not isinstance(self.index, SimIndex) and type(self.index_units) is str:
            self.index = SimIndex(self.index, units=self.index_units)
        elif type(self.index_units) is str:
            self.index.set_units(self.index_units)

    def transpose(self):
        return self

    def slope(self, x=None, y=None, window=None, slope=True, intercept=False):
        """
        calculates the slope of the series vs its index.

        Calculates the slope of column Y vs column X or vs index if 'x' is None

        Parameters
        ----------
        window : int, float or str, optional
            The half-size of the rolling window to calculate the slope.
            if None : the slope will be calculated from the entire dataset.
            if int : window rows before and after of each row will be used to calculate the slope
            if float : the window size will be variable, with window values of X arround each row's X. Not compatible with datetime columns
            if str : the window string will be used as timedelta arround the datetime X
            The default is None.
        slope : bool, optional
            Set it True to return the slope of the linear fit. The default is True.
        intercept : bool, optional
            Set it True to return the intersect of the linear fit. The default is False.
        if both slope and intercept are True, a tuple of both results will be returned

        x : kept for compatibility with SimDataFrame
        y : kept for compatibility with SimDataFrame

        Returns
        -------
        numpy array
            The array containing the desired output.

        """
        if window is None and x is not None and y is None:
            window, x = x, None
        params_ = self.params_.copy()
        if self.name is not None and len(self.get_units(self.name)) == 1 and self.index_units is not None:
            if type(params_['units']) is dict:
                params_['units'][self.name] = str(self.get_units(self.name)[self.name]) + '/' + str(self.index_units)
            else:
                params_['units'] = str(self.get_units(self.name)[self.name]) + '/' + str(self.index_units)
        params_['name'] = 'slope_of_' + (self.name)
        slope_series= _slope(df=self, x=x, y=y, window=window, slope=slope, intercept=intercept)
        return SimSeries(data=slope_series, index=self.index, **params_)

    def sort_values(self, axis=0, ascending=True, inplace=False, kind='quicksort',
                    na_position='last', ignore_index=False, key=None):
        if inplace:
            super().sort_values(axis=axis, ascending=ascending, inplace=inplace,
                                kind=kind, na_position=na_position, ignore_index=ignore_index, key=key)
            return None
        else:
            return SimSeries(
                data=self.as_Series().sort_values(axis=axis, ascending=ascending,
                                                  inplace=inplace, kind=kind,
                                                  na_position=na_position,
                                                  ignore_index=ignore_index,
                                                  key=key),
                **self.params_)

    @property
    def type(self):
        return 'SimSeries'

    def plot(self, y=None, x=None, others=None, **kwargs):
        """
        wrapper of Pandas plot method, with some superpowers

        Parameters
        ----------
        y : string, list or index; optional
            column name to plot. The default is None.
        x : string, optional
            the columns to be used for x coordinates. The default is the index.
        others : SimDataFrame, SimSeries, DataFrame or Series; optional
            other Frames to include in the plot, for the same selected columns. The default is None.
        **kwargs : TYPE
            any other keyword argument for matplolib.

        Returns
        -------
        matplotlib AxesSubplot.
        """
        return self.sdf.plot(y=y, x=x, others=others, **kwargs)



# -*- coding: utf-8 -*-
"""
Created on Sun Oct 11 11:14:32 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.80.4'
__release__ = 20230112
__all__ = ['clean_axis', 'string_new_name', 'type_of_frame', 'main_key', 'item_key']


import pandas as pd


def clean_axis(axis=None):
    if axis is None:
        return 0
    if type(axis) is str and axis.lower() in ['row', 'rows', 'ind', 'index']:
        return 0
    if type(axis) is str and axis.lower() in ['col', 'cols', 'column', 'columns']:
        return 1
    if type(axis) is str and axis.lower() in ['both', 'rows&cols', 'cols&rows', 'all']:
        return 2
    if type(axis) is bool:
        return int(axis)
    if type(axis) is float:
        return int(axis)
    return axis


def string_new_name(new_name, intersection_character='∩'):
    if len(new_name) == 1:
        return list(new_name.values())[0]
    else:
        return intersection_character.join(map(str, dict.fromkeys(new_name.values())))


def type_of_frame(frame):
    from simpandas import SimSeries, SimDataFrame
    from pandas import Series, DataFrame
    try:
        if frame._isSimSeries:
            return SimSeries
    except:
        try:
            if frame._isSimDataFrame:
                return SimDataFrame
        except:
            if type(frame) is Series:
                return Series
            elif type(frame) is DataFrame:
                return DataFrame
            else:
                raise TypeError('frame is not an instance of Pandas or SimPandas frames')


def main_key(Key, clean=True, nameSeparator=':'):
    """
    returns the main part (before the name of the item) in the keyword,MAIN:ITEM
    """
    if type(Key) is str:
        if len(Key.strip()) > 0:
            return Key.strip().split(nameSeparator)[0]
        else:
            return ''
    if type(Key) is tuple and len(Key) == 2:
        return main_key(str(Key[0]), clean=clean, nameSeparator=nameSeparator)
    if type(Key) is list or type(Key) is tuple:
        results = []
        for K in Key:
            results.append(main_key(K))
        if clean:
            return list(set(results))
        else:
            return list(results)
    if isinstance(Key, pd.Series):
        return main_key(Key.name)


def item_key(Key, clean=True, nameSeparator=':'):
    """
    returns the item part (after the name of the item) in the keyword, MAIN:ITEM
    """
    if type(Key) is str:
        if len(Key.strip()) > 0:
            if nameSeparator in Key.strip():
                return Key.strip().split(nameSeparator)[-1]
        else:
            return ''
    if type(Key) is tuple and len(Key) == 2:
        return item_key(str(Key[0]), clean=clean, nameSeparator=nameSeparator)
    if type(Key) is list or type(Key) is tuple:
        results = []
        for K in Key:
            results.append(item_key(K))
        if clean:
            return list(set(results))
        else:
            return list(results)
    if isinstance(Key, pd.Series):
        return item_key(Key.name)



# -*- coding: utf-8 -*-
"""
Created on Sun Jan 23 22:38:16 2022

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.5.1'
__release__ = 20220920
__all__ = ['slope']

import numpy as np
import pandas as pd
import warnings


def slope(df, x=None, y=None, window=None, slope=True, intercept=False):
    """
    Calculates the slope of column Y vs column X or vs index if 'x' is None

    Parameters
    ----------
    df : DataFrame or SimDataFrame
        The DataFrame to work with.
    x : str, optional
        The name of the column to be used as X.
        If None, the index of the DataFrame will be used as X.
        The default is None.
    y : str, optional
        The name of the column to be used as Y.
        If None, the first argument will be considered as Y (not as X).
        The default is None.
    window : int, float or str, optional
        The half-size of the rolling window to calculate the slope.
        if None : the slope will be calculated from the entire dataset.
        if int : window rows before and after of each row will be used to calculate the slope
        if float : the window size will be variable, with window values of X arround each row's X. Not compatible with datetime columns
        if str : the window string will be used as timedelta arround the datetime X
        The default is None.
    slope : bool, optional
        Set it True to return the slope of the linear fit. The default is True.
    intercept : bool, optional
        Set it True to return the intersect of the linear fit. The default is False.
    if both slope and intercept are True, a tuple of both results will be returned

    Returns
    -------
    numpy array
        The array containing the desired output.

    """
    # def _deltaDate(datetime):
    #     return np.cumsum(np.array([0] + list(datetime.astype('timedelta64[s]').astype(float)/60/60/24)))

    # def _checkDateTime(data):
    #     if isinstance(data, pd.Series):
    #         return _deltaDate(data)
    #     if 'datetime'

    warnings.simplefilter('ignore', np.RankWarning)

    # understanding input paramenters
    df = df.squeeze()
    if isinstance(df, (pd.Series,)):
        if window is None and y is None and x is not None:
            window, x = x, None
    elif x is not None and y is None and window is None:
        if x not in df.columns:
            window, x = x, None
    elif x is not None and y is not None and window is None and y not in df.columns:
        y, window = None, y


    # if window is None, calculate the slope for the entire dataset
    if window is None:

        if x is None:
            if isinstance(df.index, pd.DatetimeIndex):
                #xx = _deltaDate(df.index)
                xx = np.cumsum(np.array([0] + list(np.diff(df.index).astype('timedelta64[s]').astype(float)/60/60/24)))
            else:
                xx = df.index.to_numpy()
        else:
            if 'datetime' in str(df[x].dtype):
                #xx = _deltaDate(df[x])
                xx = np.cumsum(np.array([0] + list(np.diff(df[x]).astype('timedelta64[s]').astype(float)/60/60/24)))
            else:
                xx = df[x].to_numpy()

        if y is None:
            if isinstance(df, (pd.Series,)):
                if 'datetime' in str(df.dtype):
                    #yy = _deltaDate(df)
                    yy = np.cumsum(np.array([0] + list(np.diff(df).astype('timedelta64[s]').astype(float)/60/60/24)))
                else:
                    yy = df.to_numpy()
            elif x is not None and x in df.columns:
                if len(df.columns) == 2:
                    dfs = df.drop(columns=x).squeeze()
                    if 'datetime' in str(dfs.dtype):
                        #yy = _deltaDate(dfs)
                        yy = np.cumsum(np.array([0] + list(np.diff(dfs).astype('timedelta64[s]').astype(float)/60/60/24)))
                    else:
                        yy = dfs.to_numpy()
                    del dfs
                elif len(df.columns) == 1:
                    dfs = df.squeeze()
                    if 'datetime' in str(dfs.dtype):
                        #yy = _deltaDate(dfs)
                        yy = np.cumsum(np.array([0] + list(np.diff(dfs).astype('timedelta64[s]').astype(float)/60/60/24)))
                    else:
                        yy = dfs.to_numpy()
                    del dfs
                else:
                    yy = df.drop(columns=x).to_numpy()
            else:
                if len(df.columns) == 1:
                    dfs = df.squeeze()
                    if 'datetime' in str(dfs.dtype):
                        #yy = _deltaDate(dfs)
                        yy = np.cumsum(np.array([0] + list(np.diff(dfs).astype('timedelta64[s]').astype(float)/60/60/24)))
                    else:
                        yy = dfs.to_numpy()
                    del dfs
                else:
                    yy = df.to_numpy()
        else:
            if len(df[y]) == 1:
                dfs = df.squeeze()
                if 'datetime' in str(dfs.dtype):
                    #yy = _deltaDate(dfs)
                    yy = np.cumsum(np.array([0] + list(np.diff(dfs).astype('timedelta64[s]').astype(float)/60/60/24)))
                else:
                    yy = dfs.to_numpy()
                del dfs


        if slope and intercept:
            return np.polyfit(xx, yy, 1, full=False)
        elif slope and not intercept:
            return np.polyfit(xx, yy, 1, full=False)[0]
        elif not slope and intercept:
            return np.polyfit(xx, yy, 1, full=False)[1]
        else:
            return None


    # calculate slope every N (window) rows
    elif type(window) is int:

        if x is None:
            if isinstance(df.index, pd.DatetimeIndex):
                #xx = [np.cumsum(np.array([0] + list(np.diff(df.index[max(0,ii-window):min(len(df),ii+window)]).astype('timedelta64[s]').astype(float)/60/60/24))) for ii in range(len(df))]
                xx = [np.cumsum(np.diff(np.array(list(df.index[0:1]) + list(df.index[max(0,ii-window):min(len(df),ii+window)]))).astype('timedelta64[s]').astype(float)/60/60/24) for ii in range(len(df))]
            else:
                xx = [df.iloc[max(0,ii-window):min(ii+window,len(df))].index for ii in range(len(df))]
        else:
            if 'datetime' in str(df[x].dtype):
                #xx = [np.cumsum(np.array([0] + list(np.diff(df.iloc[max(0,ii-window):min(len(df),ii+window)][x]).astype('timedelta64[s]').astype(float)/60/60/24))) for ii in range(len(df))]
                xx = [np.cumsum(np.diff(np.array(list(df.iloc[0:1]) + list(df.iloc[max(0,ii-window):min(len(df),ii+window)][x]))).astype('timedelta64[s]').astype(float)/60/60/24) for ii in range(len(df))]
            else:
                xx = [df.iloc[max(0,ii-window):min(ii+window,len(df))][x] for ii in range(len(df))]

        if y is None:
            if isinstance(df, (pd.Series,)):
                yy = [df.iloc[max(0,ii-window):min(len(df),ii+window)].to_numpy() for ii in range(len(df))]
            elif x is not None and x in df.columns:
                dfs = df.drop(columns=x).squeeze()
                yy = [dfs.iloc[max(0,ii-window):min(len(dfs),ii+window)].to_numpy() for ii in range(len(dfs))]
            else:
                yy = [df.iloc[max(0,ii-window):min(len(df),ii+window)].to_numpy() for ii in range(len(df))]
        else:
            if isinstance(df, (pd.Series,)):
                yy = [df.iloc[max(0,ii-window):min(len(df),ii+window)].to_numpy() for ii in range(len(df))]
            else:
                yy = [df.iloc[max(0,ii-window):min(len(df),ii+window)][y].to_numpy() for ii in range(len(df))]


    # window is a str representing a timedelta
    elif type(window) is str:
        if ((x is None and isinstance(df.index,pd.DatetimeIndex)) or (x is not None and 'datetime' in str(df[x].dtype))):
            window = pd.to_timedelta(window)
        else:
            raise ValueError('string window, representing timedelta, only works with datetime index or X column.')

        if x is None:
            xx = [np.cumsum(np.diff(np.array(list(df.index[0:1]) + list(df[(df.index>=(df.index[ii]-window)) & (df.index<=(df.index[ii]+window))].index))).astype('timedelta64[s]').astype(float)/60/60/24) for ii in range(len(df))]
        else:
            xx = [np.cumsum(np.diff(np.array(list(df.iloc[0:1][x]) + list(df[(df.index>=(df.index[ii]-window)) & (df.index<=(df.index[ii]+window))][x]))).astype('timedelta64[s]').astype(float)/60/60/24) for ii in range(len(df))]

        if y is None:
            if isinstance(df, (pd.Series,)):
                yy = [df[(df.index>=(df.index[ii]-window)) & (df.index<=(df.index[ii]+window))].to_numpy() for ii in range(len(df))]
            elif x is not None and x in df.columns:
                dfs = df.drop(columns=x).squeeze()
                yy = [dfs[(dfs.index>=(dfs.index[ii]-window)) & (dfs.index<=(dfs.index[ii]+window))].to_numpy() for ii in range(len(dfs))]
            else:
                yy = [df[(df.index>=(df.index[ii]-window)) & (df.index<=(df.index[ii]+window))].to_numpy() for ii in range(len(df))]
        else:
            if isinstance(df, (pd.Series,)):
                yy = [df[(df.index>=(df.index[ii]-window)) & (df.index<=(df.index[ii]+window))].to_numpy() for ii in range(len(df))]
            else:
                yy = [df[(df.index>=(df.index[ii]-window)) & (df.index<=(df.index[ii]+window))][y].to_numpy() for ii in range(len(df))]


    else:
        raise NotImplementedError("'window' must be int, float, str or None.")

    polys = np.array([np.polyfit(xx[ii], yy[ii], 1, full=False) for ii in range(len(xx))])

    warnings.simplefilter('default', np.RankWarning)

    if slope and intercept:
        return polys
    elif slope and not intercept:
        return polys[:,0]
    elif not slope and intercept:
        return polys[:,1]


# -*- coding: utf-8 -*-
"""
Created on Sun Jan 08 11:10:15 2023

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.81.0'
__release__ = 20230108
__all__ = ['zeros']


def zeros(series_or_frame, axis=None, value=0):
    """
    Finds the columns or rows where all its values are equal to `value` (by default is 0).
    Returns a Series with bool indicating True at the columns or rows that match the condition.

    Parameters
    ----------
    axis = int or str
        0 for rows,
        1 for columns

    Returns
    -------
        Series
    """
    if axis is None:
        if hasattr(series_or_frame, 'columns'):
            axis = 1 if axis is None and len(series_or_frame.columns) == 1 else 0
        else:
            axis = 1
    axis = _clean_axis(axis)
    if axis == 2:
        return zeros(axis=0, value=value) + zero(axis=1, value=value)
    if hasattr(series_or_frame, 'columns'):
        limit = len(series_or_frame) if axis == 0 else len(series_or_frame.columns)
    else:
        limit = len(series_or_frame)
    return (series_or_frame == value).sum(axis=axis) == limit


def key_to_string(filter_str, key, pandas_agg):
    if len(key) > 0:
        # catch particular operations performed by Pandas
        found_so, found_no = '', ''
        if key in special_operation:
            if filter_str[-1] == ' ':
                filter_str = filter_str.rstrip()
            filter_str += key + '()'
        else:
            for SO in special_operation:
                if key.strip().endswith(SO):
                    key = key[:-len(SO)]
                    found_so = (SO if SO != '.null' else '.isnull') + '()'
                    break
        # catch particular operations performed by Numpy
        if key in numpy_operation:
            raise ValueError("wrong syntax for '" + key + "(blank space before) in:\n   " + conditions)
        else:
            for no in numpy_operation:
                if key.strip().endswith(no):
                    key = key[:-len(no)]
                    no = '.log' if no == '.ln' else no
                    filter_str += 'np' + no + '('
                    found_no = ' )'
                    break
        # catch aggregation operations performed by Pandas
        if key in pandas_aggregation:
            pandas_agg = key + '(axis=1)'
        else:
            for pa in pandas_aggregation:
                if key.strip().endswith(pa):
                    pandas_agg = pa + '(axis=1)'
                    break
        # if key is the index
        if key in ['.i', '.index']:
            filter_str = filter_str.rstrip()
            filter_str += ' self.as_pandas().index'
        # if key is a column
        elif key in self.columns:
            filter_str = filter_str.rstrip()
            filter_str += " self.as_pandas()['" + key + "']"
        # key might be a wellname, attribute or a pattern
        elif len(self.find_Keys(key)) == 1:
            filter_str = filter_str.rstrip()
            filter_str += " self.as_pandas()['" + self.find_Keys(key)[0] + "']"
        elif len(self.find_Keys(key)) > 1:
            filter_str = filter_str.rstrip()
            filter_str += " self.as_pandas()[" + str(list(self.find_Keys(key))) + "]"
            pandas_agg = '.any(axis=1)'
        else:
            filter_str += ' ' + key
        filter_str = filter_str.rstrip()
        filter_str += found_so + found_no
        key = ''
    return filter_str, key, pandas_agg



# -*- coding: utf-8 -*-
"""
Created on Sun Oct 11 11:14:32 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.80.2'
__release__ = 20230104
__all__ = ['jitter', 'znorm', 'minmaxnorm']

import numpy as np


def jitter(df, std=0.10):
    jit = np.random.randn(len(df), len(df.columns))
    jit = (jit * std) + 1
    return df * jit


def znorm(df):
    return (df - df.mean()) / df.std()


def minmaxnorm(df):
    return (df - df.min()) / (df.max() - df.min())



# -*- coding: utf-8 -*-
"""
Created on Wed Sep 18 12:33:46 2019

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.15.3'
__release__ = 20230104
__all__ = ['multisplit', 'is_numeric', 'get_number', 'is_date', 'date']

from simpandas.errors import UndefinedDateFormatError

import numpy as np
import pandas as pd
import datetime as dt


def multisplit(string, sep=[' '], remove=[' ']) :
    """
    receives a string and returns a list with string split
    by all the separators in sep.
    the default separator is the blank space ' '.
    use the remove parameter to indicate the separators that
    must not be reported in the output list.
    by default, the blank space is not reported.
    """
    assert type(string) is str

    # check sep is list
    if type(sep) is str :
        sep = [sep]

    # eliminate duplicated separators
    sep = list( set( sep ) )

    # sort sep by lenght
    s = len(sep)
    for i in range(s-1) :
        for j in range(s-i-1) :
            if len(sep[j]) < len(sep[j+1]) :
                sep[j], sep[j+1] = sep[j+1], sep[j]

    # initialize counters
    stringlist = []
    i, x, t = 0, 0, len(string)
    # loop through the entire string
    while i < t :
        found = False # flag for found separator
        # look for each separator
        for se in sep :
            s = len(se)
            if (i+s <= t) and string[i:i+s] == se :
                stringlist += [ string[x:i], se ]
                x = i+s
                i += s
                found = True
                break
        i += 1 if not found else 0
    stringlist += [ string[x:] ]

    # clean the output
    newlist = []
    for part in stringlist :
        if part not in remove + [''] :
            newlist += [ part ]

    return newlist


def is_numeric(string):
    """
    returns True if the string is a number
    """
    try:
        complex(string)
        return True
    except ValueError:
        if type(string) is str:
            string = string.strip()
            if ' ' in string:
                string = string.replace(' ', '')
            if "'" in string:
                string = string.replace("'", "")
            if ',' in string and '.' not in string:
                string = string.replace(',', '.')
            if ',' in string and '.' in string:
                string = string.replace('.', '').replace(',', '.')
            if string.startswith('(') and string.endswith(')') and len(string) > 2:
                string = '-' + string[1:-1]
            try:
                complex(string)
                return True
            except:
                return False
        else:
            return False


def get_number(string):
    "returns the number, as integer or float, contained in a string"

    def cast_to_best_number(string):
        """
        helper funtion to attempt to convert the imput to int, to float or to complex
        """
        try:
            return int(string)
        except ValueError:
            try:
                return float(string)
            except ValueError:
                try:
                    return complex(string)
                except ValueError:
                    return None

    def clean_string(string):
        """
        helper function to clean the input string
        """
        if type(string) is str:
            string = string.strip()
            if ' ' in string:
                string = string.replace(' ', '')
            if "'" in string:
                string = string.replace("'", "")
            if ',' in string and '.' not in string:
                string = string.replace(',', '.')
            if ',' in string and '.' in string:
                string = string.replace('.', '').replace(',', '.')
            if string.startswith('(') and string.endswith(')') and len(string) > 2:
                string = '-' + string[1:-1]
            return string

    attempt = cast_to_best_number(string)
    if attempt is None:
        attempt = cast_to_best_number(clean_string(string))
    if attempt is None:
        raise ValueError("could not convert string to number: " + str(string))
    else:
        return attempt


def is_date(dateStr, formatIN='', speak=False, returnFormat=False ):
    """
    returns True if the string 'dateStr' is a valid date, otherwise returns False.
    """
    if formatIN != '' :
        try :
            date(dateStr, formatIN=formatIN, speak=speak, returnFormat=returnFormat )
            if returnFormat :
                return date(dateStr, formatIN=formatIN, speak=speak, returnFormat=True )
            return True
        except :
            return False

    else :
        try :
            date(dateStr, formatIN='', speak=speak )
            if returnFormat :
                return date(dateStr, formatIN='', speak=speak, returnFormat=True )
            return True
        except :
            pass

        formats = [ 'DD-MM-YYYY', 'DD-MMM-YYYY', 'YYYY-MM-DD', 'YYYY-MMM-DD', 'MM-DD-YYYY', 'MMM-DD-YYYY', 'YYYY-DD-MM', 'YYYY-DD-MMM', 'YYYYMMDD', 'YYYYMMMDD', 'DD-MM-YY', 'MMM-DD-YY', 'MM-DD-YY'  ]
        separators = ['-', '/', ' ', '\t', '_', ':', ';', ', ', '.', '#', "'"]
        for f in formats :
            for sep in separators :
                fIN = f.replace('-', sep) if sep != '-' else f
                try :
                    date(dateStr, formatIN=fIN, speak=speak )
                    if returnFormat :
                        return fIN
                    return True
                except :
                    pass

        formats = [ 'YYYYMMDD', 'YYYYMMMDD' ]
        for f in formats :
            try :
                date(dateStr, formatIN=fIN, speak=speak )
                if returnFormat :
                    return f
                return True
            except :
                pass
    return False

def splitDMMMY(string) :
    mi, mf = len(string)+1, len(string)+1
    for x in range(len(string)) :
        if not string[x].isdigit() and x < mi:
            mi = x
        if string[x].isdigit() and x < mf and x > mi:
            mf = x
            break
    if mi > 0 and mf > 0 :
        return [string[:mi], string[mi:mf], string[mf:]]

def date(date, formatIN='', formatOUT='', speak=True, YYbaseIN=1900, returnFormat=False ):
    """
    stringformat.date receives a string containing a date or a list of strings
    containing dates and changes the date format to the format especified by
    the user. By default the out format will be 'DD-MMM-YYYY'.

    The input and output format can be stated with the keywords formatIN
    and formatOUT followed by a string containing the characters 'D', 'M'
    and 'Y' to identify day, month and year and the characters '/', '-', ' ',
    '\t' (tab) or '_' as separators.

    If the keyword formatIN is not entered, the program will try to infer
    the date format from the provided data.

    syntax examples:

    stringformat.date('31/DEC/1984', formatIN='DD/MMM/YYYY', formatOUT='MM-DD-YYYY')

    speak parameter set to True will print a message showing the input and output formats.
    """
    npDateOnly = lambda x : x.split('T')[0]

    MonthString2Number = {'JAN' :  1,
                          'FEB' :  2,
                          'MAR' :  3,
                          'APR' :  4,
                          'MAY' :  5,
                          'JUN' :  6,
                          'JLY' :  7,
                          'JUL' :  7,
                          'AUG' :  8,
                          'SEP' :  9,
                          'OCT' : 10,
                          'NOV' : 11,
                          'DEC' : 12 }
    MonthNumber2String = dict(zip( MonthString2Number.values(), MonthString2Number.keys()  ))

    separator = '' # initialize
    formatIN = formatIN.upper().strip()
    formatOUT = formatOUT.upper().strip()
    # define if input is a list/tuple of dates or a single date
    sample = str(date)
    output = list
    if type(date) is list or type(date) is tuple:
        output = list
        if type(date[0]) is str :
            for i in range(len(date)) :
                date[i] = date[i].strip()
            sample = date[0].strip()
        elif type(date[0]) is np.datetime64 :
            date = np.array(date)
        elif type(date[0]) is np.str_ :
            date = list( map( str, date ) )
            sample = date[0].strip()

    if type(date) is pd.Series :
        date = date.to_numpy()

    if type(date) is np.ndarray :
        output = list
        sample = date[0]
        if 'datetime64' in str(date.dtype) :
            date = list(np.datetime_as_string(date))
            date = list(map( npDateOnly, date ))
            formatIN = 'YYYY-MM-DD'
            separator = '-'

    if type(date) is np.datetime64 :
        formatIN = 'YYYY-MM-DD'
        date = np.datetime_as_string(date)
        date = npDateOnly( date )
        sample = date
        output = str
        separator = '-'

    if type(date) is pd.Timestamp : # pd._libs.tslibs.timestamps.Timestamp
        date = date.date()

    if type(date) is dt.date :
        date = str( date )
        if formatIN == '' :
            formatIN = 'YYYY-MM-DD'

    if type(date) is dt.datetime :
        date = str( date ).split()[0]
        if formatIN == '' :
            formatIN = 'YYYY-MM-DD'

    if type(date) is str :
        sample = date.strip(' "\'')
        date = [ date ]
        output = str

    # look for the separator, empty string if not found
    if separator == '' :
        for sep in ['/', '-', ' ', '\t', '_', ':', ';', ', ', '.', '#', "'"] :
            if sep in sample :
                separator = sep
                break

    # separate the 1st, 2nd and 3rd components of the DATEs in three lists
    if separator != '' :
        # separate the 1st, 2nd and 3rd components of the DATEs in three lists
        datelist = separator.join(date).split(separator)
        datelist = [ datelist[0::3], datelist[1::3], datelist[2::3] ]

    else :
        l = 0
        if max(map(len, date)) == min(map(len, date)) :
            l = max(map(len, date))

        if formatIN != '' :
            x, y = 0, 0
            for i in range(1, len(formatIN)) :
                if formatIN[i] != formatIN[i-1] :
                    if x == 0 :
                        x = i
                    else :
                        y = i
                        break
            datelist = [ [d[:x], d[x:y], d[y:]] for d in date ]
            datelist = [ [ datelist[i][0] for i in range(len(datelist)) ], [ datelist[i][1] for i in range(len(datelist)) ], [ datelist[i][2] for i in range(len(datelist)) ] ]

        elif l == 6 :
            datelist = [ [d[0:2], d[2:4], d[4:6]] for d in date ]
            datelist = [ [ datelist[i][0] for i in range(len(datelist)) ], [ datelist[i][1] for i in range(len(datelist)) ], [ datelist[i][2] for i in range(len(datelist)) ] ]
        elif l == 8 :
            datelist = [ [d[0:2], d[2:4], d[4:8]] for d in date ]
            datelist = [ [ datelist[i][0] for i in range(len(datelist)) ], [ datelist[i][1] for i in range(len(datelist)) ], [ datelist[i][2] for i in range(len(datelist)) ] ]
            if int(max(datelist[0])) <= 31 and int(min(datelist[2])) >= 1900 and int(max(datelist[2])) <= 2050 and int(max(datelist[1]))<=12 :
                pass # DDMMYYYY
            else :
                datelist = [ [d[0:4], d[4:6], d[6:8]] for d in date ]
                datelist = [ [ datelist[i][0] for i in range(len(datelist)) ], [ datelist[i][1] for i in range(len(datelist)) ], [ datelist[i][2] for i in range(len(datelist)) ] ]
                if int(max(datelist[2])) <= 31 and int(min(datelist[0])) >= 1900 and int(max(datelist[0])) <= 2050 and int(max(datelist[1]))<=12 :
                    pass # YYYYMMDD
                else :
                    raise UndefinedDateFormatError('unable to idenfy date format, please provide with keyword formatIN')
        elif l == 9 :
            x, y = 0, 0
            for i in range(9) :
                if not date[0][i].isdigit() and x == 0 :
                    x = i
                elif date[0][i].isdigit() and x > 0 :
                    y = i
                    break
            datelist = [ [d[:x], d[x:y], d[y:]] for d in date ]
            datelist = [ [ datelist[i][0] for i in range(len(datelist)) ], [ datelist[i][1] for i in range(len(datelist)) ], [ datelist[i][2] for i in range(len(datelist)) ] ]
        else :
            raise UndefinedDateFormatError('unable to idenfy date format, please provide with keyword formatIN')


    # if formatIN is not defined try to guess what it is
    if formatIN == '' :
        datestr = [False, False, False]
        datemax = [None, None, None]

        for i in range(3) :
            for j in range(len(date)) :
                try:
                    datelist[i][j] = int(datelist[i][j])
                except :
                    datestr[i] = True
                    break
            if datestr[i] == False :
                datemax[i] = max(datelist[i])

        orderIN = [None, None, None, separator, None, None, None]
        found = ''
        if True in datestr :
            orderIN[5] = 3
            found = found + 'Ms'
        for i in range(3) :
            if datestr[i] == True :
                orderIN[1] = i
                found = found + 'M'
            elif datemax[i] != None and datemax[i] > 999 :
                orderIN[2] = i
                orderIN[6] = 4
                found = found + 'Y'
            elif datemax[i] != None and datemax[i] > 99 :
                orderIN[2] = i
                orderIN[6] = 3
                found = found + 'Y'
            elif datemax[i] != None and datemax[i] > 31 :
                orderIN[2] = i
                orderIN[6] = 2
                found = found + 'Y'
            elif datemax[i] != None and datemax[i] > 12 and datemax[i] < 32 :
                orderIN[0] = i
                orderIN[4] = 2
                found = found + 'D'
            else :
                pass

        if None in orderIN :
            for i in range(3) :
                if datemax[i] != None and datemax[i] <= 12 :
                    if 'D' in found and 'M' not in found:
                        orderIN[1] = i
                        orderIN[5] = 2
                        found = found + 'M'
                    elif 'M' in found and 'D' not in found :
                        orderIN[0] = i
                        orderIN[4] = 2
                        found = found + 'D'

        if 'Ms' in found :
            found = found[2:]

        if 'D' in found and 'M' in found and 'Y' in found :
            formatIN = []
            for i in range(3) :
                if orderIN[i] == 0 :
                    formatIN.append('D'*orderIN[4])
                elif orderIN[i] == 1 :
                    formatIN.append('M'*orderIN[5])
                elif orderIN[i] == 2 :
                    formatIN.append('Y'*orderIN[6])
            formatIN = orderIN[3].join(formatIN)
            if speak :
                print(' the input format is: ' + formatIN)

        else :
            raise UndefinedDateFormatError('unable to idenfy date format, please provide with keyword formatIN')

        if returnFormat :
            return formatIN

    # read input format from formatIN
    else :
        orderIN = [None, None, None, None, None, None, None] # [day, month, year, separator, day_digit, month_digits, year_digits]
        for sep in ['/', '-', ' ', '\t', '_', ':', ';', '#', "'"] :
            if sep in formatIN :
                orderIN[3] = sep
                break
        indexDMY = [ formatIN.upper().index('D'), formatIN.upper().index('M'), formatIN.upper().index('Y') ]
        for i in range(3) :
            if indexDMY[i] == min(indexDMY):
                orderIN[i] = 0
            elif indexDMY[i] == max(indexDMY):
                orderIN[i] = 2
            else :
                orderIN[i] = 1
        orderIN[4] = formatIN.upper().count('D')
        orderIN[5] = formatIN.upper().count('M')
        orderIN[6] = formatIN.upper().count('Y')

        for sep in ['/', '-', ' ', '\t'] :
            if sep in formatIN :
                test = sep
                break


    # set formatOUT by default if not provided
    if formatOUT == '' :
        formatOUT = 'DD-MMM-YYYY'
        orderOUT = [0, 1, 2, '-', 2, 3, 4]
        # if speak and formatIN != formatOUT :
        #     print(' default output format is: DD-MMM-YYYY')

    # read format from formatOUT
    else :
        orderOUT = [None, None, None, '', None, None, None] # [day, month, year, separator, day_digit, month_digits, year_digits]
        for sep in ['/', '-', ' ', '\t', '_', ':', ';', '#', "'"] :
            if sep in formatOUT :
                orderOUT[3] = sep
                break
        if 'D' in formatOUT.upper() :
            indexD = formatOUT.upper().index('D')
        else :
            indexD = 2
        if 'M' in formatOUT.upper() :
            indexM = formatOUT.upper().index('M')
        else :
            indexM = 2
        if 'Y' in formatOUT.upper() :
            indexY = formatOUT.upper().index('Y')
        else :
            indexY = 2
        indexDMY = [ indexD, indexM, indexY ]
        for i in range(3) :
            if indexDMY[i] == min(indexDMY):
                orderOUT[i] = 0
            elif indexDMY[i] == max(indexDMY):
                orderOUT[i] = 2
            else :
                orderOUT[i] = 1
        orderOUT[4] = formatOUT.upper().count('D')
        orderOUT[5] = formatOUT.upper().count('M')
        orderOUT[6] = formatOUT.upper().count('Y')


    dateOUT = [ datelist[orderIN.index(orderOUT[0])], datelist[orderIN.index(orderOUT[1])], datelist[orderIN.index(orderOUT[2])] ]

    if orderOUT[5] == 0 :
        dateM = ''
    elif orderOUT[5] == 5 :
        dateM = orderOUT[1]
        for i in range(len(dateOUT[dateM])) :
            dateOUT[dateM][i] = str(int(dateOUT[dateM][i])).zfill(2) + MonthNumber2String[int(dateOUT[dateM][i])]
    elif orderOUT[5] > 2 and orderIN[5] <= 2 :
        dateM = orderOUT[1]
        for i in range(len(dateOUT[dateM])) :
            dateOUT[dateM][i] = MonthNumber2String[int(dateOUT[dateM][i])]
    elif orderOUT[5] <= 2 and orderIN[5] > 2 :
        dateM = orderOUT[1]
        for i in range(len(dateOUT[dateM])) :
            dateOUT[dateM][i] = MonthString2Number[dateOUT[dateM][i]]

    dateOUTformated = []
    numberformat = [None, None, None] # [year, day, month]
    for i in range(3) :
        numberformat[orderOUT[i]] = orderOUT[i+4]
    for i in range(len(dateOUT[0])) :
        #print(numberformat)
        if numberformat[0] == 0 or numberformat[0] == None:
            dateStr = ''
        elif type(dateOUT[0][i]) == int and numberformat[0] == 2 and dateOUT[0][i] < 10 :
            dateStr = '0' + str(dateOUT[0][i]) + orderOUT[3]
        elif type(dateOUT[0][i]) == int and numberformat[0] == 3 and dateOUT[0][i] < 10 :
            dateStr = '00' + str(dateOUT[0][i]) + orderOUT[3]
        elif type(dateOUT[0][i]) == int and numberformat[0] == 3 and dateOUT[0][i] < 100 :
            dateStr = '0' + str(dateOUT[0][i]) + orderOUT[3]
        elif type(dateOUT[0][i]) == int and numberformat[0] == 4 and dateOUT[0][i] < 10 :
            if YYbaseIN == 0 :
                dateStr = '000' + str(dateOUT[0][i]) + orderOUT[3]
            else :
                dateStr = str(dateOUT[0][i]+YYbaseIN) + orderOUT[3]
        elif type(dateOUT[0][i]) == int and numberformat[0] == 4 and dateOUT[0][i] < 100 :
            if YYbaseIN == 0 :
                dateStr = '00' + str(dateOUT[0][i]) + orderOUT[3]
            else :
                dateStr = str(dateOUT[0][i]+YYbaseIN) + orderOUT[3]
        elif type(dateOUT[0][i]) == int and numberformat[0] == 4 and dateOUT[0][i] < 1000 :
            dateStr = '0' + str(dateOUT[0][i]) + orderOUT[3]
        else :
            dateStr = str(dateOUT[0][i]) + orderOUT[3]

        if numberformat[1] == 0 or numberformat[1] == None :
            dateStr = dateStr + ''
        elif type(dateOUT[1][i]) == int and numberformat[1] == 2 and dateOUT[1][i] < 10 :
            dateStr = dateStr + '0' + str(dateOUT[1][i]) + orderOUT[3]
        elif type(dateOUT[1][i]) == int and numberformat[1] == 3 and dateOUT[1][i] < 10 :
            dateStr = dateStr + '00' + str(dateOUT[1][i]) + orderOUT[3]
        elif type(dateOUT[1][i]) == int and numberformat[1] == 3 and dateOUT[1][i] < 100 :
            dateStr = dateStr + '0' + str(dateOUT[1][i]) + orderOUT[3]
        elif type(dateOUT[1][i]) == int and numberformat[1] == 4 and dateOUT[1][i] < 10 :
            if YYbaseIN == 0 :
                dateStr = dateStr + '000' + str(dateOUT[1][i]) + orderOUT[3]
            else :
                dateStr = dateStr + str(dateOUT[1][i]+YYbaseIN) + orderOUT[3]
        elif type(dateOUT[1][i]) == int and numberformat[1] == 4 and dateOUT[1][i] < 100 :
            if YYbaseIN :
                dateStr = dateStr + '00' + str(dateOUT[1][i]) + orderOUT[3]
            else :
                dateStr = dateStr + str(dateOUT[1][i]+YYbaseIN) + orderOUT[3]
        elif type(dateOUT[1][i]) == int and numberformat[1] == 4 and dateOUT[1][i] < 1000 :
            dateStr = dateStr + '0' + str(dateOUT[1][i]) + orderOUT[3]
        else :
            dateStr = dateStr + str(dateOUT[1][i]) + orderOUT[3]

        if numberformat[2] == 0 or numberformat[2] == None:
            dateStr = dateStr + ''
        elif type(dateOUT[2][i]) == int and numberformat[2] == 2 and dateOUT[2][i] < 10 :
            dateStr = dateStr + '0' + str(dateOUT[2][i])
        elif type(dateOUT[2][i]) == int and numberformat[2] == 3 and dateOUT[2][i] < 10 :
            dateStr = dateStr + '00' + str(dateOUT[2][i])
        elif type(dateOUT[2][i]) == int and numberformat[2] == 3 and dateOUT[2][i] < 100 :
            dateStr = dateStr + '0' + str(dateOUT[2][i])
        elif type(dateOUT[2][i]) == int and numberformat[2] == 4 and dateOUT[2][i] < 10 :
            if YYbaseIN == 0 :
                dateStr = dateStr + '000' + str(dateOUT[2][i])
            else :
                dateStr = dateStr + str(dateOUT[2][i]+YYbaseIN)
        elif type(dateOUT[2][i]) == int and numberformat[2] == 4 and dateOUT[2][i] < 100 :
            if YYbaseIN == 0 :
                dateStr = dateStr + '00' + str(dateOUT[2][i])
            else :
                dateStr = dateStr + str(dateOUT[2][i]+YYbaseIN)
        elif type(dateOUT[2][i]) == int and numberformat[2] == 4 and dateOUT[2][i] < 1000 :
            dateStr = dateStr + '0' + str(dateOUT[2][i])
        else :
            dateStr = dateStr + str(dateOUT[2][i])

        dateOUTformated.append( dateStr )

    if output is str :
        return dateOUTformated[0]
    else :
        return dateOUTformated